"""Core generation functions for ai-query."""

from __future__ import annotations

import asyncio
import inspect
from typing import Any, AsyncIterator

from ai_query.types import (
    GenerateTextResult,
    Message,
    ProviderOptions,
    TextPart,
    ToolCallPart,
    ToolResultPart,
    Usage,
    TextStreamResult,
    StreamChunk,
    ToolSet,
    ToolCall,
    ToolResult,
    StepResult,
    StopCondition,
    step_count_is,
    StepStartEvent,
    StepFinishEvent,
    OnStepStart,
    OnStepFinish,
    EmbedResult,
    EmbedManyResult,
    EmbeddingUsage,
    AbortSignal,
    AbortError,
)
from ai_query.model import LanguageModel, EmbeddingModel


async def generate_text(
    *,
    model: LanguageModel,
    prompt: str | None = None,
    system: str | None = None,
    messages: list[Message] | list[dict[str, Any]] | None = None,
    tools: ToolSet | None = None,
    stop_when: StopCondition | list[StopCondition] | None = None,
    on_step_start: OnStepStart | None = None,
    on_step_finish: OnStepFinish | None = None,
    provider_options: ProviderOptions | None = None,
    signal: AbortSignal | None = None,
    **kwargs: Any,
) -> GenerateTextResult:
    """Generate text using an AI model."""
    # Build messages list
    final_messages: list[Message] = []

    if messages is not None:
        for msg in messages:
            if isinstance(msg, Message):
                final_messages.append(msg)
            else:
                final_messages.append(Message(role=msg["role"], content=msg["content"]))
    else:
        if system:
            final_messages.append(Message(role="system", content=system))
        if prompt:
            final_messages.append(Message(role="user", content=prompt))

    if not final_messages:
        raise ValueError("Either 'prompt' or 'messages' must be provided")

    stop_conditions: list[StopCondition] = []
    if stop_when is not None:
        if isinstance(stop_when, list):
            stop_conditions = stop_when
        else:
            stop_conditions = [stop_when]

    if not stop_conditions:
        stop_conditions = [step_count_is(1)]

    steps: list[StepResult] = []
    current_messages = list(final_messages)
    total_usage = Usage()
    accumulated_text = ""
    final_result: GenerateTextResult | None = None
    step_number = 0

    if signal:
        signal.throw_if_aborted()

    while True:
        step_number += 1

        if signal and signal.aborted:
            raise AbortError(signal.reason)

        if on_step_start:
            start_event = StepStartEvent(
                step_number=step_number,
                messages=current_messages,
                tools=tools,
            )
            callback_result = on_step_start(start_event)
            if inspect.isawaitable(callback_result):
                await callback_result

        result = await model.provider.generate(
            model=model.model_id,
            messages=current_messages,
            tools=tools,
            provider_options=provider_options,
            **kwargs,
        )

        if result.usage:
            total_usage.input_tokens += result.usage.input_tokens
            total_usage.output_tokens += result.usage.output_tokens
            total_usage.cached_tokens += result.usage.cached_tokens
            total_usage.total_tokens += result.usage.total_tokens

        tool_calls: list[ToolCall] = result.response.get("tool_calls", [])

        step = StepResult(
            text=result.text,
            tool_calls=tool_calls,
            tool_results=[],
            finish_reason=result.finish_reason,
        )

        if result.text:
            accumulated_text += result.text

        if not tool_calls:
            final_result = result
            final_result.usage = total_usage
            steps.append(step)

            if on_step_finish:
                finish_event = StepFinishEvent(
                    step_number=step_number,
                    step=step,
                    text=accumulated_text,
                    usage=total_usage,
                    steps=steps,
                )
                callback_result = on_step_finish(finish_event)
                if inspect.isawaitable(callback_result):
                    await callback_result

            break

        tool_results: list[ToolResult] = []
        for tc in tool_calls:
            tool_def = tools.get(tc.name) if tools else None
            if tool_def is None:
                tool_results.append(ToolResult(
                    tool_call_id=tc.id,
                    tool_name=tc.name,
                    result=f"Error: Unknown tool '{tc.name}'",
                    is_error=True,
                ))
                continue

            try:
                output = await tool_def.run(**tc.arguments)
                tool_results.append(ToolResult(
                    tool_call_id=tc.id,
                    tool_name=tc.name,
                    result=output,
                ))
            except Exception as e:
                tool_results.append(ToolResult(
                    tool_call_id=tc.id,
                    tool_name=tc.name,
                    result=f"Error: {e}",
                    is_error=True,
                ))

        step.tool_results = tool_results
        steps.append(step)

        if on_step_finish:
            finish_event = StepFinishEvent(
                step_number=step_number,
                step=step,
                text=accumulated_text,
                usage=total_usage,
                steps=steps,
            )
            callback_result = on_step_finish(finish_event)
            if inspect.isawaitable(callback_result):
                await callback_result

        should_stop = False
        for condition in stop_conditions:
            cond_result = condition(steps)
            if inspect.isawaitable(cond_result):
                cond_result = await cond_result
            if cond_result:
                should_stop = True
                break

        if should_stop:
            final_result = result
            final_result.usage = total_usage
            break

        assistant_content: list = []
        if result.text:
            assistant_content.append(TextPart(text=result.text))
        for tc in tool_calls:
            assistant_content.append(ToolCallPart(tool_call=tc))

        current_messages.append(Message(
            role="assistant",
            content=assistant_content if assistant_content else "",
        ))

        tool_result_parts = [ToolResultPart(tool_result=tr) for tr in tool_results]
        current_messages.append(Message(
            role="tool",
            content=tool_result_parts,
        ))

    if final_result:
        final_result.steps = steps
        return final_result

    raise RuntimeError("Generation loop ended without a result")


def stream_text(
    *,
    model: LanguageModel,
    prompt: str | None = None,
    system: str | None = None,
    messages: list[Message] | list[dict[str, Any]] | None = None,
    tools: ToolSet | None = None,
    stop_when: StopCondition | list[StopCondition] | None = None,
    on_step_start: OnStepStart | None = None,
    on_step_finish: OnStepFinish | None = None,
    provider_options: ProviderOptions | None = None,
    signal: AbortSignal | None = None,
    **kwargs: Any,
) -> TextStreamResult:
    """Stream text from an AI model."""
    final_messages: list[Message] = []

    if messages is not None:
        for msg in messages:
            if isinstance(msg, Message):
                final_messages.append(msg)
            else:
                final_messages.append(Message(role=msg["role"], content=msg["content"]))
    else:
        if system:
            final_messages.append(Message(role="system", content=system))
        if prompt:
            final_messages.append(Message(role="user", content=prompt))

    if not final_messages:
        raise ValueError("Either 'prompt' or 'messages' must be provided")

    stop_conditions: list[StopCondition] = []
    if stop_when is not None:
        if isinstance(stop_when, list):
            stop_conditions = stop_when
        else:
            stop_conditions = [stop_when]

    if not stop_conditions:
        stop_conditions = [step_count_is(1)]

    shared_steps: list[StepResult] = []

    async def _stream_generator() -> AsyncIterator[StreamChunk]:
        nonlocal shared_steps
        steps: list[StepResult] = shared_steps
        current_messages = list(final_messages)
        total_usage = Usage()
        accumulated_text = ""
        step_number = 0

        if signal:
            signal.throw_if_aborted()

        while True:
            step_number += 1

            if signal and signal.aborted:
                raise AbortError(signal.reason)

            if on_step_start:
                start_event = StepStartEvent(
                    step_number=step_number,
                    messages=current_messages,
                    tools=tools,
                )
                callback_result = on_step_start(start_event)
                if inspect.isawaitable(callback_result):
                    await callback_result

            stream = model.provider.stream(
                model=model.model_id,
                messages=current_messages,
                tools=tools,
                provider_options=provider_options,
                **kwargs,
            )

            step_text = ""
            step_tool_calls: list[ToolCall] = []
            step_finish_reason = None

            async for chunk in stream:
                if chunk.is_final:
                    if chunk.usage:
                        total_usage.input_tokens += chunk.usage.input_tokens
                        total_usage.output_tokens += chunk.usage.output_tokens
                        total_usage.cached_tokens += chunk.usage.cached_tokens
                        total_usage.total_tokens += chunk.usage.total_tokens
                    step_finish_reason = chunk.finish_reason
                    if chunk.tool_calls:
                        step_tool_calls = chunk.tool_calls
                else:
                    if chunk.text:
                        if signal and signal.aborted:
                            raise AbortError(signal.reason)
                        step_text += chunk.text
                        accumulated_text += chunk.text
                        yield StreamChunk(text=chunk.text)

            step = StepResult(
                text=step_text,
                tool_calls=step_tool_calls,
                tool_results=[],
                finish_reason=step_finish_reason,
            )

            if not step_tool_calls:
                steps.append(step)

                if on_step_finish:
                    finish_event = StepFinishEvent(
                        step_number=step_number,
                        step=step,
                        text=accumulated_text,
                        usage=total_usage,
                        steps=steps,
                    )
                    callback_result = on_step_finish(finish_event)
                    if inspect.isawaitable(callback_result):
                        await callback_result

                yield StreamChunk(
                    is_final=True,
                    usage=total_usage,
                    finish_reason=step_finish_reason,
                )
                break

            tool_results: list[ToolResult] = []
            for tc in step_tool_calls:
                tool_def = tools.get(tc.name) if tools else None
                if tool_def is None:
                    tool_results.append(ToolResult(
                        tool_call_id=tc.id,
                        tool_name=tc.name,
                        result=f"Error: Unknown tool '{tc.name}'",
                        is_error=True,
                    ))
                    continue

                try:
                    output = await tool_def.run(**tc.arguments)
                    tool_results.append(ToolResult(
                        tool_call_id=tc.id,
                        tool_name=tc.name,
                        result=output,
                    ))
                except Exception as e:
                    tool_results.append(ToolResult(
                        tool_call_id=tc.id,
                        tool_name=tc.name,
                        result=f"Error: {str(e)}",
                        is_error=True,
                    ))

            step.tool_results = tool_results
            steps.append(step)

            if on_step_finish:
                finish_event = StepFinishEvent(
                    step_number=step_number,
                    step=step,
                    text=accumulated_text,
                    usage=total_usage,
                    steps=steps,
                )
                callback_result = on_step_finish(finish_event)
                if inspect.isawaitable(callback_result):
                    await callback_result

            should_stop = False
            for condition in stop_conditions:
                cond_result = condition(steps)
                if inspect.isawaitable(cond_result):
                    cond_result = await cond_result
                if cond_result:
                    should_stop = True
                    break

            if should_stop:
                yield StreamChunk(
                    is_final=True,
                    usage=total_usage,
                    finish_reason=step_finish_reason,
                )
                break

            assistant_content: list = []
            if step_text:
                assistant_content.append(TextPart(text=step_text))
            for tc in step_tool_calls:
                assistant_content.append(ToolCallPart(tool_call=tc))

            current_messages.append(Message(
                role="assistant",
                content=assistant_content if assistant_content else "",
            ))

            tool_result_parts = [ToolResultPart(tool_result=tr) for tr in tool_results]
            current_messages.append(Message(
                role="tool",
                content=tool_result_parts,
            ))

    return TextStreamResult(_stream_generator(), steps=shared_steps)


async def embed(
    *,
    model: EmbeddingModel,
    value: str,
    provider_options: ProviderOptions | None = None,
    signal: AbortSignal | None = None,
    **kwargs: Any,
) -> EmbedResult:
    """Generate an embedding for a single value."""
    if signal:
        signal.throw_if_aborted()

    if signal:
        embed_task = asyncio.create_task(
            model.provider.embed(
                model=model.model_id,
                value=value,
                provider_options=provider_options,
                **kwargs,
            )
        )
        abort_task = asyncio.create_task(signal.wait())

        done, pending = await asyncio.wait(
            [embed_task, abort_task],
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()

        if abort_task in done:
            raise AbortError(signal.reason)

        return embed_task.result()
    else:
        return await model.provider.embed(
            model=model.model_id,
            value=value,
            provider_options=provider_options,
            **kwargs,
        )


async def embed_many(
    *,
    model: EmbeddingModel,
    values: list[str],
    provider_options: ProviderOptions | None = None,
    signal: AbortSignal | None = None,
    **kwargs: Any,
) -> EmbedManyResult:
    """Generate embeddings for multiple values."""
    if signal:
        signal.throw_if_aborted()

    tasks = [
        embed(model=model, value=v, provider_options=provider_options, signal=signal, **kwargs)
        for v in values
    ]
    results = await asyncio.gather(*tasks)

    total_tokens = sum(r.usage.tokens for r in results)
    return EmbedManyResult(
        values=values,
        embeddings=[r.embedding for r in results],
        usage=EmbeddingUsage(tokens=total_tokens),
    )



from __future__ import annotations

import asyncio
from typing import Any

from aiohttp import web

from ai_query.agents import (
    Agent,
    action,
    AgentServer,
    MemoryStorage,
    ConnectionContext,
    Connection,
)


class ChatRoom(Agent):
    enable_event_log = True
    def __init__(self, id: str, **kwargs: Any):
        super().__init__(
            id,
            storage=MemoryStorage(),
            initial_state={"messages": [], "reputation": {}},
            **kwargs
        )
    

    async def on_connect(self, connection: Connection, ctx: ConnectionContext):
        """Set user_id and notify room of arrival."""
        user_id = ctx.request.headers.get("X-User-Id", "anonymous")
        connection.state["user_id"] = user_id
        await self.emit("status", {"text": f"{user_id} has joined."})
        await super().on_connect(connection, ctx)

    async def on_message(self, connection: Connection, message: str | bytes | dict[str, Any]):
        """Broadcast incoming chat messages to all users."""
        if isinstance(message, (bytes, str)):
            text = message.decode("utf-8") if isinstance(message, bytes) else message
        elif isinstance(message, dict):
            text = message.get("text", "")
        else:
            return

        user_id = connection.state.get("user_id", "anonymous")
        message_data = {"user_id": user_id, "text": text}
        
        # Add to history
        self.state["messages"].append(message_data)
        
        # Broadcast to all clients
        await self.emit("message", message_data)

    @action
    async def add_reaction(self, message_id: str, reaction: str) -> dict[str, Any]:
        """Add a reaction to a message."""
        assert self.context.connection is not None
        user_id = self.context.connection.state.get("user_id", "anonymous")
        await self.emit("reaction", {"user_id": user_id, "reaction": reaction})
        return {"status": "ok", "user": user_id}

    @action
    async def get_history(self, user_id: str | None = None) -> list[dict[str, Any]]:
        """Get the chat history, optionally filtering by user_id."""
        if user_id:
            return [
                msg for msg in self.state["messages"] if msg.get("user_id") == user_id
            ]
        return self.state["messages"]

    @action
    async def greet(self, name: str) -> str:
        """A simple greeting action."""
        return f"Hello, {name}!"


if __name__ == "__main__":
    server = AgentServer(ChatRoom)
    server.serve(port=8080)



"""Code Executor Agent

Demonstrates:
- Executing Python code in a sandboxed environment
- Tool with potentially dangerous operations (handled safely)
- Step callbacks for monitoring execution
"""

import asyncio
import io
from contextlib import redirect_stdout, redirect_stderr
from ai_query import generate_text, tool, Field, step_count_is, StepStartEvent, StepFinishEvent
from ai_query.providers import google


# --- Code Execution Tool ---

@tool(description=(
    "Execute Python code and return the output. "
    "Use print() to display results. "
    "Available: math module, basic builtins (len, range, list, dict, sum, max, min, sorted, etc.)"
))
def execute_python(
    code: str = Field(description="The Python code to execute")
) -> str:
    """Execute Python code and return the output."""
    print(f"  [Executor] Running code...")

    # Capture stdout and stderr
    stdout_capture = io.StringIO()
    stderr_capture = io.StringIO()

    # Restricted globals for safety
    safe_globals = {
        "__builtins__": {
            "print": print,
            "len": len,
            "range": range,
            "list": list,
            "dict": dict,
            "set": set,
            "tuple": tuple,
            "str": str,
            "int": int,
            "float": float,
            "bool": bool,
            "sum": sum,
            "max": max,
            "min": min,
            "abs": abs,
            "round": round,
            "sorted": sorted,
            "enumerate": enumerate,
            "zip": zip,
            "map": map,
            "filter": filter,
            "all": all,
            "any": any,
            "isinstance": isinstance,
            "type": type,
            "True": True,
            "False": False,
            "None": None,
        },
        "math": __import__("math"),
    }

    local_vars = {}

    try:
        with redirect_stdout(stdout_capture), redirect_stderr(stderr_capture):
            exec(code, safe_globals, local_vars)

        stdout = stdout_capture.getvalue()
        stderr = stderr_capture.getvalue()

        result = ""
        if stdout:
            result += f"Output:\n{stdout}"
        if stderr:
            result += f"\nStderr:\n{stderr}"
        if not result:
            # Check if there's a result variable or last expression
            if "result" in local_vars:
                result = f"Result: {local_vars['result']}"
            else:
                result = "Code executed successfully (no output)"

        return result.strip()

    except Exception as e:
        return f"Error: {type(e).__name__}: {e}"


# --- Callbacks ---

def on_start(event: StepStartEvent):
    print(f"\n[Step {event.step_number}] Starting...")


def on_finish(event: StepFinishEvent):
    if event.step.tool_calls:
        for tc in event.step.tool_calls:
            print(f"[Step {event.step_number}] Executing code:")
            code = tc.arguments.get("code", "")
            for line in code.split("\n"):
                print(f"    {line}")

    if event.step.tool_results:
        for tr in event.step.tool_results:
            print(f"[Step {event.step_number}] Result: {tr.result}")


# --- Main ---

async def main():
    print("Python Code Executor Agent")
    print("=" * 50)

    problems = [
        "Calculate the first 10 Fibonacci numbers",
        "Find all prime numbers between 1 and 50",
        "Calculate the factorial of 12",
    ]

    for problem in problems:
        print(f"\nProblem: {problem}")
        print("-" * 40)

        result = await generate_text(
            model=google("gemini-2.0-flash"),
            system=(
                "You are a Python coding assistant. When asked to solve problems, "
                "write and execute Python code to find the answer. "
                "Always use print() to display your results clearly."
            ),
            prompt=problem,
            tools={"execute_python": execute_python},
            on_step_start=on_start,
            on_step_finish=on_finish,
            stop_when=step_count_is(4),
        )

        print(f"\nAnswer: {result.text}")
        print(f"Tokens: {result.usage.total_tokens}")
        print("=" * 50)


if __name__ == "__main__":
    asyncio.run(main())



"""Connection adapters for AgentServer."""

from __future__ import annotations

import json
from typing import Any

from aiohttp import web

from ai_query.agents.websocket import Connection


class AioHttpConnection(Connection):
    """Wraps aiohttp WebSocket in our Connection interface."""
    state: dict[str, Any]

    def __init__(self, ws: web.WebSocketResponse, request: web.Request):
        self._ws = ws
        self._request = request
        self.username: str | None = None
        self.agent_id: str | None = None
        self.state = {}

    async def send(self, message: str | bytes) -> None:
        if isinstance(message, bytes):
            await self._ws.send_bytes(message)
        else:
            await self._ws.send_str(message)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        await self._ws.close(code=code, message=reason.encode())

    async def send_event(self, event: str, data: dict[str, Any]) -> None:
        """Send a structured event as JSON."""
        payload = {"type": event, **data}
        await self.send(json.dumps(payload))


class AioHttpSSEConnection(Connection):
    """Wraps aiohttp StreamResponse for SSE in our Connection interface."""
    state: dict[str, Any]

    def __init__(self, response: web.StreamResponse, request: web.Request):
        self._response = response
        self._request = request
        self.state = {}

    async def send(self, message: str | bytes) -> None:
        """Send message as SSE data event."""
        # This handles raw send calls (e.g. broadcast)
        # We assume it's a JSON string, wrap it in 'message' event or generic data
        if isinstance(message, bytes):
            message = message.decode("utf-8")

        # If message looks like JSON with 'type', try to use that?
        # Simpler: just send as data
        await self._response.write(f"data: {message}\n\n".encode("utf-8"))

    async def send_event(self, event: str, data: dict[str, Any]) -> None:
        """Send structured event as SSE."""
        # Escape newlines for SSE data
        json_data = json.dumps(data)
        await self._response.write(f"event: {event}\ndata: {json_data}\n\n".encode("utf-8"))

    async def close(self, code: int = 1000, reason: str = "") -> None:
        # SSE connections are closed by client disconnecting or server finishing
        # We can't explicitly close from here except by finishing response
        await self._response.write_eof()



import asyncio
import json
import re
from typing import Any, Dict, List, Optional, Type, Union

# Try to import workers.DurableObject (Cloudflare runtime)
try:
    from workers import DurableObject
except ImportError:

    class DurableObject:  # type: ignore
        def __init__(self, ctx: Any, env: Any):
            self.ctx = ctx
            self.env = env


try:
    import js
    from pyodide.ffi import to_js, create_proxy
except ImportError:
    js = None
    to_js = lambda x: x
    create_proxy = lambda x: x

from ai_query.agents.agent import Agent
from ai_query.agents.storage.cloudflare import DurableObjectStorage
from ai_query.agents.transport.cloudflare import DurableObjectTransport
from ai_query.agents.websocket import Connection, ConnectionContext


class WebSocketBridge(Connection):
    """Bridges a Cloudflare WebSocket to the Agent Connection protocol."""

    def __init__(self, ws: Any):
        self._ws = ws

    async def send(self, data: Union[str, bytes]) -> None:
        """Send data to the client."""
        # Check if ws is still open? Cloudflare WS object handles this.
        self._ws.send(data)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        """Close the connection."""
        self._ws.close(code, reason)


class AgentDO(DurableObject):
    """Base class for Cloudflare Durable Objects hosting an Agent.

    Subclass this and set `agent_class` to your Agent implementation.
    Inherits from `workers.DurableObject`.

    Implements WebSocket Hibernation API: https://developers.cloudflare.com/durable-objects/best-practices/websockets/#websocket-hibernation-api
    """

    agent_class: Type[Agent]

    def __init__(self, ctx: Any, env: Any):
        super().__init__(ctx, env)
        self.ctx = ctx
        self.env = env

        # The `id` is available on the context object in the new Python runtime
        self.id = str(ctx.id) if hasattr(ctx, "id") else "unknown"

        self.storage = DurableObjectStorage(ctx.storage)
        self.transport = DurableObjectTransport(env)

        # Provider options extraction
        provider_options: Dict[str, Any] = {}
        for key in [
            "OPENAI_API_KEY",
            "ANTHROPIC_API_KEY",
            "GEMINI_API_KEY",
            "OPENROUTER_API_KEY",
        ]:
            val = getattr(env, key, None)
            if val:
                provider_options[key] = val

        self.agent = self.agent_class(
            id=self.id,
            storage=self.storage,
            transport=self.transport,
            provider_options=provider_options,
        )
        if hasattr(self.agent_class, "model") and not self.agent.model:
            self.agent.model = self.agent_class.model
        self.agent.env = env  # type: ignore

        # Inject emit handler for WebSocket and SSE delivery
        async def cloudflare_emit_handler(
            event_type: str, data: Dict[str, Any], event_id: int
        ) -> None:
            payload = json.dumps({"id": event_id, "type": event_type, "data": data})

            # 1. Broadcast to all accepted WebSockets (Hibernation API)
            for ws in self.ctx.getWebSockets():
                try:
                    ws.send(payload)
                except Exception:
                    pass

            # 2. Broadcast to all active SSE connections
            # We convert to SSE format: "event: <type>\ndata: <json>\n\n"
            sse_payload = f"event: {event_type}\ndata: {json.dumps(data)}\n\n"
            for sse_queue in self.agent._sse_connections:
                try:
                    sse_queue.put_nowait(sse_payload)
                except Exception:
                    pass

        self.agent._emit_handler = cloudflare_emit_handler

    async def fetch(self, request: Any) -> Any:
        try:
            # Lazy start: Ensure agent is initialized
            if self.agent._state is None:
                await self.agent.start()

            headers = request.headers
            upgrade = headers.get("Upgrade")
            if upgrade == "websocket":
                return self.handle_websocket_upgrade(request)

            if request.method == "POST":
                body = await request.json()
                data = body.to_py() if hasattr(body, "to_py") else body
                result = await self.agent.handle_request(data)

                # Ensure background tasks (like emit) are completed before hibernation
                self.ctx.waitUntil(self._drain_mailbox())

                return self._json_response(result)
            elif request.method == "GET":
                return self._json_response({"agent_id": self.agent.id})

            return js.Response.new("Method not allowed", {"status": 405})

        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error in DO fetch: {e}")
            return self._json_response({"error": str(e)}, status=500)

    def handle_websocket_upgrade(self, request: Any) -> Any:
        """Handle WebSocket upgrade using Hibernation API."""
        # Use Pythonic way to unpack JS iterable
        client, server = js.WebSocketPair.new().object_values()

        # Accept the connection. This attaches the WebSocket to the Durable Object.
        self.ctx.acceptWebSocket(server)

        bridge = WebSocketBridge(server)
        # We assume root path for connection context as standard convention
        ctx = ConnectionContext(path="/", headers={}, query_params={})

        # Trigger the initial on_connect event asynchronously
        # We use waitUntil to ensure on_connect finishes (e.g. if it emits events)
        connect_task = asyncio.create_task(self.agent.on_connect(bridge, ctx))
        self.ctx.waitUntil(connect_task)

        return js.Response.new(
            None,
            js.Object.fromEntries(to_js({"status": 101, "webSocket": client})),
        )

    # --- WebSocket Hibernation Events ---
    # These methods are called by the runtime when events occur on accepted WebSockets.

    async def webSocketMessage(self, ws: Any, message: Any) -> None:
        """Called when a message is received from a WebSocket."""
        # Ensure agent is running. If the DO was hibernating, this will reload state.
        if not self.agent._running:
            await self.agent.start()

        bridge = WebSocketBridge(ws)
        # message might be string or bytes
        await self.agent.on_message(bridge, message)

        # Ensure any side-effects from processing the message are completed
        self.ctx.waitUntil(self._drain_mailbox())

    async def webSocketClose(
        self, ws: Any, code: int, reason: str, wasClean: bool
    ) -> None:
        """Called when a WebSocket is closed."""
        bridge = WebSocketBridge(ws)
        await self.agent.on_close(bridge, code, reason)

    async def webSocketError(self, ws: Any, error: Any) -> None:
        """Called when a WebSocket error occurs."""
        # Errors are typically handled by the connection closing or client retrying
        pass

    def _json_response(self, data: Any, status: int = 200) -> Any:
        return js.Response.new(
            json.dumps(data),
            js.Object.fromEntries(
                to_js(
                    {"status": status, "headers": {"Content-Type": "application/json"}}
                )
            ),
        )

    async def _drain_mailbox(self) -> None:
        """Wait until the agent's mailbox is empty."""
        # join() blocks until all items in the queue have been processed and task_done() is called.
        await self.agent._mailbox.join()

    async def alarm(self) -> None:
        if not self.agent._processor_task or self.agent._processor_task.done():
            self.agent._processor_task = asyncio.create_task(
                self.agent._process_mailbox()
            )


class CloudflareRegistry:
    def __init__(self, env: Any):
        self.env = env
        self.routes: List[tuple[re.Pattern, Any]] = []

    def register(self, pattern: str, binding: Any) -> None:
        self.routes.append((re.compile(pattern), binding))

    async def handle_request(self, request: Any) -> Any:
        url = js.URL.new(request.url)
        path = url.pathname

        parts = path.strip("/").split("/")
        if len(parts) >= 2 and parts[0] == "agent":
            agent_id = parts[1]
            binding = None
            for pattern, b in self.routes:
                if pattern.match(agent_id):
                    binding = b
                    break

            if binding:
                stub = binding.getByName(agent_id)
                return await stub.fetch(request)

        return js.Response.new(
            "Agent not found", js.Object.fromEntries(to_js({"status": 404}))
        )



"""Main entry point for ai-query library."""

from __future__ import annotations

from ai_query.types import (
    GenerateTextResult,
    TextStreamResult,
    EmbedResult,
    EmbedManyResult,
    tool,
    Field,
    StepStartEvent,
    StepFinishEvent,
    OnStepStart,
    OnStepFinish,
    StopCondition,
    step_count_is,
    has_tool_call,
)
from ai_query.model import LanguageModel, EmbeddingModel
from ai_query.core import (
    generate_text,
    stream_text,
    embed,
    embed_many,
)

from ai_query.agents.agent import Agent, Event, action
from ai_query.agents.registry import AgentRegistry
from ai_query.agents.remote import connect
from ai_query.agents.server import AgentServer, AgentServerConfig
from ai_query.agents.transport.http import HTTPTransport
from ai_query.agents.transport import LocalTransport

__all__ = [
    # Main functions
    "generate_text",
    "stream_text",
    "embed",
    "embed_many",
    # Agents
    "Agent",
    "Event",
    "action",
    "AgentRegistry",
    "AgentServer",
    "AgentServerConfig",
    "connect",
    "HTTPTransport",
    "LocalTransport",
    # Tool decorators
    "tool",
    "Field",
    # Common types
    "LanguageModel",
    "EmbeddingModel",
    "GenerateTextResult",
    "TextStreamResult",
    "EmbedResult",
    "EmbedManyResult",
    # Stop conditions
    "StopCondition",
    "step_count_is",
    "has_tool_call",
    # Callbacks
    "StepStartEvent",
    "StepFinishEvent",
    "OnStepStart",
    "OnStepFinish",
]



import asyncio
import argparse
import aiohttp
import json
import sys


async def monitor_websocket(url: str, agent_id: str):
    """Connects to the WebSocket and prints incoming messages."""
    ws_url = f"{url}/agent/{agent_id}"
    ws_url = ws_url.replace("https://", "wss://").replace("http://", "ws://")

    print(f"Connecting to WebSocket: {ws_url} ...")

    try:
        async with aiohttp.ClientSession() as session:
            async with session.ws_connect(ws_url) as ws:
                print("✓ Connected! Waiting for updates...")
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        data = json.loads(msg.data)
                        print(f"\n[WebSocket] Received: {json.dumps(data, indent=2)}")
                        if data.get("type", "") == "error":
                            print("Error received, exiting.")
                            break
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        print("ws connection closed with exception %s", ws.exception())
    except Exception as e:
        print(f"WebSocket error: {e}")


async def interactive_loop(url: str, agent_id: str):
    """Sends HTTP requests to increment or chat."""
    invoke_url = f"{url}/agent/{agent_id}/invoke"
    chat_url = f"{url}/agent/{agent_id}/chat"

    print(f"\n[Commands] 'i': increment, 'q': quit, anything else: chat")

    async with aiohttp.ClientSession() as session:
        while True:
            cmd = await asyncio.to_thread(input, "> ")
            if cmd.lower() == "q":
                break

            try:
                if cmd.lower() == "i":
                    # Increment Action
                    payload = {"payload": {"method": "increment"}}
                    async with session.post(invoke_url, json=payload) as resp:
                        if resp.status != 200:
                            print(f"Error {resp.status}: {await resp.text()}")
                            continue
                        data = await resp.json()
                        print(f"[Count] New Value: {data.get('result')}")
                else:
                    # Chat Message
                    payload = {"message": cmd}
                    async with session.post(chat_url, json=payload) as resp:
                        if resp.status != 200:
                            print(f"Error {resp.status}: {await resp.text()}")
                            continue
                        data = await resp.json()
                        print(f"[AI] {data.get('response')}")

            except Exception as e:
                print(f"Request failed: {e}")


async def main():
    parser = argparse.ArgumentParser(description="Cloudflare Counter Client")
    parser.add_argument(
        "url",
        help="The base URL of your worker (e.g. https://my-worker.user.workers.dev)",
    )
    parser.add_argument("--id", default="counter-1", help="The Agent ID to target")
    args = parser.parse_args()

    # Run WebSocket monitor in background
    ws_task = asyncio.create_task(monitor_websocket(args.url.rstrip("/"), args.id))

    # Run interactive loop
    await interactive_loop(args.url.rstrip("/"), args.id)

    # Cleanup
    ws_task.cancel()
    try:
        await ws_task
    except asyncio.CancelledError:
        pass


if __name__ == "__main__":
    try:
        if len(sys.argv) < 2:
            print("Usage: python client.py <WORKER_URL>")
            print(
                "Example: python client.py https://ai-query-counter.my-user.workers.dev"
            )
            sys.exit(1)
        asyncio.run(main())
    except KeyboardInterrupt:
        print("\nGoodbye!")




import asyncio
import aiohttp
import argparse
import uuid

async def main(user_id: str):
    """Connect to an agent's SSE endpoint and print events."""
    agent_id = "test-room"
    url = f"http://localhost:8080/agent/{agent_id}/events?last_event_id=0"
    
    print(f"Connecting to SSE event stream for agent '{agent_id}' as user '{user_id}'...")
    print(f"URL: {url}")

    try:
        async with aiohttp.ClientSession() as session:
            headers = {"X-User-Id": user_id, "Accept": "text/event-stream"}
            async with session.get(url, headers=headers) as response:
                if response.status != 200:
                    print(f"Error connecting: {response.status}")
                    body = await response.text()
                    print(body)
                    return

                print("Connection successful. Listening for server events...")
                # The connection will stay open and receive events indefinitely
                while True:
                    line = await response.content.readline()
                    if not line:
                        break
                    line_str = line.decode('utf-8').strip()
                    if line_str:
                        print(f"<<< {line_str}")

    except aiohttp.ClientConnectorError as e:
        print(f"Connection error: {e}")
        print("Is the agent server running? Try: `python examples/rpc/server.py`")
    except asyncio.CancelledError:
        print("\nClient shut down.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Agent SSE Event Stream Client")
    parser.add_argument(
        "--user-id",
        type=str,
        default=f"sse-user-{uuid.uuid4().hex[:6]}",
        help="Unique ID for the client user.",
    )
    args = parser.parse_args()

    try:
        asyncio.run(main(args.user_id))
    except KeyboardInterrupt:
        print("\nClient shut down.")



"""OpenRouter provider - OpenAI-compatible API."""

from __future__ import annotations

import os

from ai_query.model import LanguageModel
from ai_query.providers.openai.provider import OpenAIProvider


class OpenRouterProvider(OpenAIProvider):
    """OpenRouter provider - wraps OpenAI provider with OpenRouter's base URL."""

    name = "openrouter"

    def __init__(self, api_key: str | None = None, **kwargs):
        resolved_api_key = api_key or os.environ.get("OPENROUTER_API_KEY")
        super().__init__(
            api_key=resolved_api_key,
            base_url="https://openrouter.ai/api/v1",
            **kwargs,
        )


# Cached provider instance
_default_provider: OpenRouterProvider | None = None


def openrouter(
    model_id: str,
    *,
    api_key: str | None = None,
) -> LanguageModel:
    """Create an OpenRouter language model.

    OpenRouter provides access to many AI models through a single API.
    Uses the OpenAI-compatible API at https://openrouter.ai/api/v1

    Args:
        model_id: The model identifier (e.g., "openai/gpt-4o", "anthropic/claude-3.5-sonnet").
        api_key: OpenRouter API key. Falls back to OPENROUTER_API_KEY env var.

    Returns:
        A LanguageModel instance for use with generate_text().

    Example:
        >>> from ai_query import generate_text, openrouter
        >>> result = await generate_text(
        ...     model=openrouter("openai/gpt-4o"),
        ...     prompt="Hello!"
        ... )
    """
    global _default_provider

    # Create provider with custom settings, or reuse default
    if api_key:
        provider = OpenRouterProvider(api_key=api_key)
    else:
        if _default_provider is None:
            _default_provider = OpenRouterProvider()
        provider = _default_provider

    return LanguageModel(provider=provider, model_id=model_id)



"""Base provider interface for ai-query."""

from __future__ import annotations

import asyncio
import json
from abc import ABC, abstractmethod
from typing import Any, AsyncIterator, TYPE_CHECKING

from ai_query.types import (
    GenerateTextResult,
    Message,
    ProviderOptions,
    StreamChunk,
    ToolSet,
    Usage,
    EmbedResult,
    EmbedManyResult,
    EmbeddingUsage,
)


class BaseProvider(ABC):
    """Abstract base class for AI providers.

    To create a new provider adapter:
    1. Subclass BaseProvider
    2. Implement the `generate` method
    3. Implement the `stream` method for streaming support
    """

    # Provider identifier (e.g., "openai", "anthropic", "google")
    name: str

    def __init__(self, api_key: str | None = None, **kwargs: Any):
        """Initialize the provider.

        Args:
            api_key: API key for the provider. If None, will try to read from
                     environment variable (provider-specific).
            **kwargs: Additional provider-specific configuration.
        """
        self.api_key = api_key
        self.config = kwargs

    @abstractmethod
    async def generate(
        self,
        *,
        model: str,
        messages: list[Message],
        tools: ToolSet | None = None,
        provider_options: ProviderOptions | None = None,
        **kwargs: Any,
    ) -> GenerateTextResult:
        """Generate text using the provider's API.

        Args:
            model: The model identifier (without provider prefix).
            messages: List of messages in the conversation.
            tools: Optional dict of tool definitions for tool calling.
            provider_options: Provider-specific options.
            **kwargs: Additional parameters (max_tokens, temperature, etc.).

        Returns:
            GenerateTextResult containing the generated text and metadata.
        """
        pass

    @abstractmethod
    async def stream(
        self,
        *,
        model: str,
        messages: list[Message],
        tools: ToolSet | None = None,
        provider_options: ProviderOptions | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[StreamChunk]:
        """Stream text using the provider's API.

        Args:
            model: The model identifier (without provider prefix).
            messages: List of messages in the conversation.
            provider_options: Provider-specific options.
            **kwargs: Additional parameters (max_tokens, temperature, etc.).

        Yields:
            StreamChunk objects containing text and metadata.
            The final chunk will have is_final=True with usage and finish_reason.
        """
        pass
        # Make this an async generator
        if False:
            yield StreamChunk()

    def get_provider_options(
        self, provider_options: ProviderOptions | None
    ) -> dict[str, Any]:
        """Extract options specific to this provider.

        Args:
            provider_options: Full provider options dict.

        Returns:
            Options for this specific provider, or empty dict.
        """
        if provider_options is None:
            return {}
        return provider_options.get(self.name, {})

    async def _fetch_resource_as_base64(
        self, url: str, session: Any
    ) -> tuple[str, str]:
        """Fetch a resource from URL and return as base64 with media type."""
        async with session.get(url) as resp:
            if resp.status != 200:
                raise Exception(f"Failed to fetch resource from {url}: {resp.status}")
            content_type = resp.headers.get("Content-Type", "application/octet-stream")
            # Extract just the mime type (remove charset etc)
            media_type = content_type.split(";")[0].strip()
            data_bytes = await resp.read()
            import base64

            return base64.b64encode(data_bytes).decode(), media_type

    def create_session(self) -> aiohttp.ClientSession:
        """Create an aiohttp ClientSession with environment-specific configuration.

        This handles Cloudflare Worker specific issues like disabling SSL verification
        which is required in that environment as the native 'ssl' module is not available.
        """
        import os
        import aiohttp

        connector = None
        # Check for Cloudflare/Pyodide environment
        # sys.platform is 'emscripten' in Cloudflare Workers (Pyodide)
        import sys

        if (
            sys.platform == "emscripten"
            or os.environ.get("WORKER_RUNTIME") == "cloudflare"
        ):
            connector = aiohttp.TCPConnector(ssl=False)

        return aiohttp.ClientSession(connector=connector)

    def _parse_sse_line(self, line: bytes | str) -> str | None:
        """Parse an SSE line and extract the data payload.

        Args:
            line: Raw SSE line (bytes or str).

        Returns:
            The data string if line starts with "data: ", None otherwise.
        """
        if isinstance(line, bytes):
            line = line.decode("utf-8")
        line = line.strip()
        if not line:
            return None
        if line.startswith("data: "):
            return line[6:]
        return None

    def _parse_sse_json(self, line: bytes | str) -> dict[str, Any] | None:
        """Parse an SSE line and decode the JSON payload.

        Args:
            line: Raw SSE line (bytes or str).

        Returns:
            Parsed JSON dict, or None if line is not valid SSE data or invalid JSON.
        """
        data = self._parse_sse_line(line)
        if data is None:
            return None
        if data == "[DONE]":
            return None
        try:
            return json.loads(data)
        except json.JSONDecodeError:
            return None

    def _accumulate_usage(self, total: Usage, delta: Usage) -> None:
        """Accumulate usage statistics in-place.

        Args:
            total: The running total Usage object (modified in-place).
            delta: The delta Usage object to add.
        """
        total.input_tokens += delta.input_tokens
        total.output_tokens += delta.output_tokens
        total.cached_tokens += delta.cached_tokens
        total.total_tokens += delta.total_tokens

    async def embed(
        self,
        *,
        model: str,
        value: str,
        provider_options: ProviderOptions | None = None,
        **kwargs: Any,
    ) -> EmbedResult:
        """Generate an embedding for a single value.

        Args:
            model: The embedding model identifier.
            value: The text to embed.
            provider_options: Provider-specific options.
            **kwargs: Additional parameters (dimensions, etc.).

        Returns:
            EmbedResult containing the embedding vector and metadata.

        Raises:
            NotImplementedError: If the provider doesn't support embeddings.
        """
        raise NotImplementedError(f"{self.name} provider does not support embeddings")

    async def embed_many(
        self,
        *,
        model: str,
        values: list[str],
        provider_options: ProviderOptions | None = None,
        **kwargs: Any,
    ) -> EmbedManyResult:
        """Generate embeddings for multiple values.

        Default implementation calls embed() for each value in parallel.
        Providers should override this for more efficient batch processing
        if their API supports it natively.

        Args:
            model: The embedding model identifier.
            values: List of texts to embed.
            provider_options: Provider-specific options.
            **kwargs: Additional parameters (dimensions, etc.).

        Returns:
            EmbedManyResult containing embedding vectors and metadata.
        """
        # Run all embed calls in parallel
        tasks = [
            self.embed(
                model=model,
                value=value,
                provider_options=provider_options,
                **kwargs,
            )
            for value in values
        ]
        results = await asyncio.gather(*tasks)

        # Aggregate results
        embeddings = [r.embedding for r in results]
        total_tokens = sum(r.usage.tokens for r in results)

        return EmbedManyResult(
            values=values,
            embeddings=embeddings,
            usage=EmbeddingUsage(tokens=total_tokens),
        )



"""Built-in aiohttp server for agents with WebSocket, SSE, and REST endpoints."""

from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING, Any, Any

from aiohttp import web, WSMsgType

from ai_query.agents.websocket import Connection, ConnectionContext

if TYPE_CHECKING:
    from ai_query.agents.agent import Agent


class AioHttpConnection(Connection):
    """Wraps aiohttp WebSocket in our Connection interface."""
    state: dict[str, Any]

    def __init__(self, ws: web.WebSocketResponse, request: web.Request):
        self._ws = ws
        self._request = request
        self.state = {}

    async def send(self, message: str | bytes) -> None:
        if isinstance(message, bytes):
            await self._ws.send_bytes(message)
        else:
            await self._ws.send_str(message)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        await self._ws.close(code=code, message=reason.encode())

    async def send_event(self, event: str, data: dict) -> None:
        """Send a structured event as JSON."""
        await self.send(json.dumps({"type": event, **data}))


async def websocket_handler(request: web.Request) -> web.WebSocketResponse:
    """Handle WebSocket connections for an agent."""
    agent: Agent = request.app["agent"]

    ws = web.WebSocketResponse()
    await ws.prepare(request)

    connection = AioHttpConnection(ws, request)
    ctx = ConnectionContext(
        request=request,
        metadata=dict(request.query),
    )

    # Handle replay on reconnect
    last_event_id = request.query.get("last_event_id")
    if last_event_id:
        try:
            after_id = int(last_event_id)
            async for event in agent.replay_events(after_id):
                await connection.send(json.dumps({
                    "type": event.type,
                    "id": event.id,
                    **event.data,
                }))
        except ValueError:
            pass

    # Register connection
    await agent.on_connect(connection, ctx)

    try:
        async for msg in ws:
            if msg.type == WSMsgType.TEXT:
                await agent.on_message(connection, msg.data)
            elif msg.type == WSMsgType.BINARY:
                await agent.on_message(connection, msg.data)
            elif msg.type == WSMsgType.ERROR:
                pass
    except Exception:
        pass
    finally:
        await agent.on_close(connection, 1000, "Client disconnected")

    return ws


async def sse_handler(request: web.Request) -> web.StreamResponse:
    """Handle SSE connections for event streaming."""
    agent: Agent = request.app["agent"]

    response = web.StreamResponse(
        headers={
            "Content-Type": "text/event-stream",
            "Cache-Control": "no-cache",
            "Connection": "keep-alive",
            "Access-Control-Allow-Origin": "*",
        }
    )
    await response.prepare(request)

    # Handle replay on reconnect
    last_event_id = request.query.get("last_event_id") or request.headers.get("Last-Event-ID")
    if last_event_id:
        try:
            after_id = int(last_event_id)
            async for event in agent.replay_events(after_id):
                sse_msg = f"id: {event.id}\nevent: {event.type}\ndata: {json.dumps(event.data)}\n\n"
                await response.write(sse_msg.encode())
        except ValueError:
            pass

    # Register SSE connection
    agent._sse_connections.add(response)

    try:
        # Keep alive until client disconnects
        while True:
            await asyncio.sleep(30)
            await response.write(b": keepalive\n\n")
    except (ConnectionResetError, asyncio.CancelledError):
        pass
    finally:
        agent._sse_connections.discard(response)

    return response


def add_cors_headers(response: web.Response) -> web.Response:
    """Add CORS headers to a response."""
    response.headers["Access-Control-Allow-Origin"] = "*"
    response.headers["Access-Control-Allow-Methods"] = "GET, POST, PUT, DELETE, OPTIONS"
    response.headers["Access-Control-Allow-Headers"] = "Content-Type, Authorization"
    return response


async def cors_preflight_handler(request: web.Request) -> web.Response:
    """Handle CORS preflight OPTIONS requests."""
    return add_cors_headers(web.Response())


async def chat_handler(request: web.Request) -> web.Response:
    """Handle REST chat requests.

    Streams the response internally while emitting chunks to SSE/WebSocket
    subscribers, then returns the complete response as JSON.
    """
    agent: Agent = request.app["agent"]

    try:
        body = await request.json()
        message = body.get("message", "")

        if not message:
            return add_cors_headers(web.json_response(
                {"error": "message is required"},
                status=400
            ))

        # Emit start event
        await agent.emit("chat_start", {"message": message})

        # Stream response, emitting chunks to SSE/WebSocket subscribers
        full_response = ""
        async for chunk in agent.stream(message):
            full_response += chunk
            await agent.emit("chunk", {"content": chunk})

        # Emit completion event
        await agent.emit("chat_complete", {"response": full_response})

        return add_cors_headers(web.json_response({
            "agent_id": agent.id,
            "response": full_response,
        }))

    except json.JSONDecodeError:
        return add_cors_headers(web.json_response(
            {"error": "Invalid JSON"},
            status=400
        ))
    except Exception as e:
        await agent.emit("chat_error", {"error": str(e)})
        return add_cors_headers(web.json_response(
            {"error": str(e)},
            status=500
        ))


async def state_handler(request: web.Request) -> web.Response:
    """Get agent state."""
    agent: Agent = request.app["agent"]

    return add_cors_headers(web.json_response({
        "agent_id": agent.id,
        "state": agent.state,
    }))


def _create_emit_handler(agent: "Agent"):
    """Create the emit handler that delivers events to all connected clients."""

    async def deliver(event: str, data: dict, event_id: int) -> None:
        # Deliver to WebSocket connections
        ws_msg = json.dumps({"type": event, "id": event_id, **data})
        for conn in list(agent._connections):
            try:
                await conn.send(ws_msg)
            except Exception:
                agent._connections.discard(conn)

        # Deliver to SSE connections
        sse_msg = f"id: {event_id}\nevent: {event}\ndata: {json.dumps(data)}\n\n"
        for sse in list(agent._sse_connections):
            try:
                await sse.write(sse_msg.encode())
            except Exception:
                agent._sse_connections.discard(sse)

    return deliver


def run_agent_server(
    agent: "Agent",
    host: str = "localhost",
    port: int = 8080,
) -> None:
    """Run the agent as a server (blocking).

    Creates endpoints:
        - WS   /ws           - WebSocket connection
        - GET  /events       - SSE event stream
        - POST /chat         - REST chat endpoint
        - GET  /state        - Get agent state

    Args:
        agent: The agent to serve.
        host: Host to bind to (default: localhost).
        port: Port to bind to (default: 8080).
    """
    asyncio.run(run_agent_server_async(agent, host, port))


async def run_agent_server_async(
    agent: "Agent",
    host: str = "localhost",
    port: int = 8080,
) -> None:
    """Run the agent as a server (async).

    Args:
        agent: The agent to serve.
        host: Host to bind to (default: localhost).
        port: Port to bind to (default: 8080).
    """
    # Start the agent
    await agent.start()

    # Inject emit handler for event delivery
    agent._emit_handler = _create_emit_handler(agent)

    # Create aiohttp app
    app = web.Application()
    app["agent"] = agent

    # Routes
    app.router.add_get("/ws", websocket_handler)
    app.router.add_get("/events", sse_handler)
    app.router.add_post("/chat", chat_handler)
    app.router.add_options("/chat", cors_preflight_handler)
    app.router.add_get("/state", state_handler)
    app.router.add_options("/state", cors_preflight_handler)

    # Run server
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, host, port)

    print(f"Agent '{agent.id}' serving at http://{host}:{port}")
    print(f"  WebSocket: ws://{host}:{port}/ws")
    print(f"  SSE:       http://{host}:{port}/events")
    print(f"  Chat:      POST http://{host}:{port}/chat")
    print(f"  State:     GET http://{host}:{port}/state")

    await site.start()

    # Run forever
    try:
        await asyncio.Event().wait()
    finally:
        await agent.stop()
        await runner.cleanup()



"""Tests for the Agent class."""

import json
import pytest
from unittest.mock import AsyncMock, MagicMock, patch
from ai_query.agents import (
    Agent,
    MemoryStorage,
    SQLiteStorage,
    Connection,
    ConnectionContext,
)
from ai_query.types import Message, tool
from ai_query.model import LanguageModel


class TestAgentInitialization:
    """Tests for Agent initialization and configuration."""

    def test_basic_initialization(self):
        """Agent should initialize with minimal required parameters."""
        agent = Agent("test-agent")
        assert agent.id == "test-agent"
        assert isinstance(agent.storage, MemoryStorage)

    def test_initialization_with_custom_storage(self, tmp_path):
        """Agent should accept custom storage backend."""
        storage = SQLiteStorage(str(tmp_path / "test.db"))
        agent = Agent("test-agent", storage=storage)
        assert agent.storage is storage
        storage.close()

    def test_initialization_with_initial_state(self):
        """Agent should accept initial_state parameter."""
        agent = Agent("test-agent", initial_state={"counter": 0, "name": "test"})
        assert agent._initial_state == {"counter": 0, "name": "test"}

    def test_initialization_with_system_prompt(self):
        """Agent should accept system parameter."""
        agent = Agent("test-agent", system="You are a helpful bot.")
        assert agent.system == "You are a helpful bot."

    def test_initialization_with_tools(self):
        """Agent should accept tools parameter."""

        def my_tool():
            return "result"

        agent = Agent("test-agent", tools={"my_tool": tool(my_tool)})
        assert "my_tool" in agent.tools


class TestAgentState:
    """Tests for Agent state management."""

    @pytest.fixture
    def storage(self):
        """Create a fresh MemoryStorage instance."""
        return MemoryStorage()

    @pytest.mark.asyncio
    async def test_initial_state_used_when_no_stored_state(self, storage):
        """Agent should use initial_state when storage is empty."""
        agent = Agent("test-agent", storage=storage, initial_state={"counter": 0})
        await agent.start()
        assert agent.state == {"counter": 0}
        await agent.stop()

    @pytest.mark.asyncio
    async def test_stored_state_takes_precedence(self, storage):
        """Agent should load stored state instead of initial_state."""
        await storage.set("test-agent:state", {"counter": 100})

        agent = Agent("test-agent", storage=storage, initial_state={"counter": 0})
        await agent.start()
        assert agent.state == {"counter": 100}
        await agent.stop()

    @pytest.mark.asyncio
    async def test_set_state_persists(self, storage):
        """set_state() should persist to storage."""
        agent = Agent("test-agent", storage=storage)
        await agent.start()
        await agent.set_state({"new": "state"})

        stored = await storage.get("test-agent:state")
        assert stored == {"new": "state"}
        await agent.stop()

    @pytest.mark.asyncio
    async def test_update_state_merges(self, storage):
        """update_state() should merge with existing state."""
        agent = Agent("test-agent", storage=storage, initial_state={"a": 1, "b": 2})
        await agent.start()
        await agent.update_state(b=20, c=3)

        assert agent.state == {"a": 1, "b": 20, "c": 3}
        await agent.stop()

    @pytest.mark.asyncio
    async def test_state_persists_across_instances(self, storage):
        """State should persist across agent instances."""
        agent1 = Agent("test-agent", storage=storage)
        await agent1.start()
        await agent1.set_state({"persisted": True})
        await agent1.stop()

        agent2 = Agent("test-agent", storage=storage)
        await agent2.start()
        assert agent2.state == {"persisted": True}
        await agent2.stop()

    def test_state_not_accessible_before_start(self):
        """Accessing state before start should raise RuntimeError."""
        agent = Agent("test-agent")
        with pytest.raises(RuntimeError, match="Agent not started"):
            _ = agent.state


class TestAgentMessages:
    """Tests for Agent message history."""

    @pytest.mark.asyncio
    async def test_messages_empty_initially(self):
        """Messages should be empty after start."""
        agent = Agent("test-agent", storage=MemoryStorage())
        await agent.start()
        assert agent.messages == []
        await agent.stop()

    @pytest.mark.asyncio
    async def test_messages_persist_across_instances(self):
        """Messages should persist in storage."""
        storage = MemoryStorage()

        agent1 = Agent("test-agent", storage=storage)
        await agent1.start()
        agent1._messages.append(Message(role="user", content="Hello"))
        agent1._messages.append(Message(role="assistant", content="Hi there!"))
        await storage.set(
            "test-agent:messages", [m.to_dict() for m in agent1._messages]
        )
        await agent1.stop()

        agent2 = Agent("test-agent", storage=storage)
        await agent2.start()
        assert len(agent2.messages) == 2
        assert agent2.messages[0].content == "Hello"
        await agent2.stop()

    @pytest.mark.asyncio
    async def test_clear_removes_messages(self):
        """clear() should remove all messages."""
        storage = MemoryStorage()
        agent = Agent("test-agent", storage=storage)
        await agent.start()
        agent._messages.append(Message(role="user", content="Hello"))
        await agent.clear()
        assert agent.messages == []
        await agent.stop()


class TestAgentLifecycle:
    """Tests for Agent lifecycle management."""

    @pytest.mark.asyncio
    async def test_context_manager(self):
        """Agent should work as async context manager."""
        async with Agent("test-agent", storage=MemoryStorage()) as agent:
            assert agent._running is True

    @pytest.mark.asyncio
    async def test_start_calls_on_start(self):
        """start() should call on_start hook."""
        on_start_called = []

        class MyAgent(Agent):
            async def on_start(self):
                on_start_called.append(True)

        agent = MyAgent("test", storage=MemoryStorage())
        await agent.start()
        assert on_start_called == [True]
        await agent.stop()

    @pytest.mark.asyncio
    async def test_stop_cancels_processor(self):
        """stop() should cancel the mailbox processor."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()
        assert agent._processor_task is not None

        await agent.stop()
        assert agent._processor_task is None


class TestAgentConnections:
    """Tests for Agent connection management."""

    @pytest.mark.asyncio
    async def test_on_connect_adds_connection(self):
        """on_connect should add connection to set."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        mock_conn = MagicMock(spec=Connection)
        ctx = ConnectionContext(request=None, metadata={})
        await agent.on_connect(mock_conn, ctx)

        assert mock_conn in agent._connections
        await agent.stop()

    @pytest.mark.asyncio
    async def test_on_close_removes_connection(self):
        """on_close should remove connection from set."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        mock_conn = MagicMock(spec=Connection)
        agent._connections.add(mock_conn)
        await agent.on_close(mock_conn, 1000, "Normal")

        assert mock_conn not in agent._connections
        await agent.stop()


class TestAgentChat:
    """Tests for Agent chat functionality."""

    @pytest.fixture
    def mock_model(self):
        """Create a mock LanguageModel."""
        model = MagicMock(spec=LanguageModel)
        model.provider = MagicMock()
        model.model_id = "test-model"
        return model

    @pytest.fixture
    def mock_stream_text(self):
        """Create a mock stream_text function."""

        async def mock_text_stream():
            yield "Hello "
            yield "World!"

        mock_result = MagicMock()
        mock_result.text_stream = mock_text_stream()

        with patch("ai_query.stream_text", return_value=mock_result) as mock:
            yield mock

    @pytest.mark.asyncio
    async def test_chat_adds_user_message(self, mock_stream_text, mock_model):
        """chat() should add user message to history."""
        agent = Agent("test", storage=MemoryStorage(), model=mock_model)
        await agent.start()

        await agent.chat("Hello")

        assert len(agent.messages) >= 1
        assert agent.messages[0].role == "user"
        assert agent.messages[0].content == "Hello"
        await agent.stop()

    @pytest.mark.asyncio
    async def test_chat_adds_assistant_response(self, mock_stream_text, mock_model):
        """chat() should add assistant response to history."""
        agent = Agent("test", storage=MemoryStorage(), model=mock_model)
        await agent.start()

        await agent.chat("Hello")

        assert len(agent.messages) == 2
        assert agent.messages[1].role == "assistant"
        assert agent.messages[1].content == "Hello World!"
        await agent.stop()

    @pytest.mark.asyncio
    async def test_chat_persists_messages(self, mock_stream_text, mock_model):
        """chat() should persist messages to storage."""
        storage = MemoryStorage()
        agent = Agent("test", storage=storage, model=mock_model)
        await agent.start()

        await agent.chat("Hello")

        stored = await storage.get("test:messages")
        assert stored is not None
        assert len(stored) == 2
        await agent.stop()


class TestAgentStream:
    """Tests for Agent streaming functionality."""

    @pytest.fixture
    def mock_model(self):
        """Create a mock LanguageModel."""
        model = MagicMock(spec=LanguageModel)
        model.provider = MagicMock()
        model.model_id = "test-model"
        return model

    @pytest.fixture
    def mock_stream_text(self):
        """Create a mock stream_text function."""

        async def mock_text_stream():
            yield "Hello "
            yield "World!"

        mock_result = MagicMock()
        mock_result.text_stream = mock_text_stream()

        with patch("ai_query.stream_text", return_value=mock_result) as mock:
            yield mock

    @pytest.mark.asyncio
    async def test_stream_yields_chunks(self, mock_stream_text, mock_model):
        """stream() should yield response chunks."""
        agent = Agent("test", storage=MemoryStorage(), model=mock_model)
        await agent.start()

        chunks = []
        async for chunk in agent.stream("Hello"):
            chunks.append(chunk)

        assert chunks == ["Hello ", "World!"]
        await agent.stop()

    @pytest.mark.asyncio
    async def test_stream_adds_messages(self, mock_stream_text, mock_model):
        """stream() should add messages to history."""
        agent = Agent("test", storage=MemoryStorage(), model=mock_model)
        await agent.start()

        async for _ in agent.stream("Hello"):
            pass

        assert len(agent.messages) == 2
        await agent.stop()


class TestAgentSubclassing:
    """Tests for Agent subclassing patterns."""

    @pytest.mark.asyncio
    async def test_subclass_with_defaults(self):
        """Subclass can set default model and system prompt."""

        class MyBot(Agent):
            def __init__(self, agent_id: str):
                super().__init__(
                    agent_id, system="I am a custom bot.", storage=MemoryStorage()
                )

        agent = MyBot("test")
        assert agent.system == "I am a custom bot."

    @pytest.mark.asyncio
    async def test_subclass_with_custom_hooks(self):
        """Subclass can override lifecycle hooks."""
        events = []

        class MyBot(Agent):
            async def on_start(self):
                events.append("start")

            async def on_stop(self):
                events.append("stop")

        agent = MyBot("test", storage=MemoryStorage())
        await agent.start()
        await agent.stop()

        assert events == ["start", "stop"]

    @pytest.mark.asyncio
    async def test_subclass_with_tools(self):
        """Subclass can define tools."""

        def search(query: str) -> str:
            return f"Results for: {query}"

        class SearchBot(Agent):
            def __init__(self, agent_id: str):
                super().__init__(
                    agent_id, tools={"search": tool(search)}, storage=MemoryStorage()
                )

        agent = SearchBot("test")
        assert "search" in agent.tools


class TestAgentMailbox:
    """Tests for Agent actor mailbox."""

    @pytest.mark.asyncio
    async def test_enqueue_connect_calls_on_connect(self):
        """Enqueued connect should call on_connect handler."""
        connected = []

        class MyAgent(Agent):
            async def on_connect(self, connection, ctx):
                await super().on_connect(connection, ctx)
                connected.append(connection)

        mock_conn = MagicMock(spec=Connection)
        ctx = ConnectionContext(request=None, metadata={})

        async with MyAgent("test", storage=MemoryStorage()) as agent:
            agent.enqueue("connect", None, connection=mock_conn, ctx=ctx)
            await agent._mailbox.join()

        assert mock_conn in connected

    @pytest.mark.asyncio
    async def test_enqueue_message_calls_on_message(self):
        """Enqueued message should call on_message handler."""
        messages = []

        class MyAgent(Agent):
            async def on_message(self, connection, message):
                messages.append(message)

        mock_conn = MagicMock(spec=Connection)

        async with MyAgent("test", storage=MemoryStorage()) as agent:
            agent.enqueue("message", "Hello!", connection=mock_conn)
            await agent._mailbox.join()

        assert "Hello!" in messages

    @pytest.mark.asyncio
    async def test_enqueue_close_calls_on_close(self):
        """Enqueued close should call on_close handler."""
        closed = []

        class MyAgent(Agent):
            async def on_close(self, connection, code, reason):
                await super().on_close(connection, code, reason)
                closed.append((code, reason))

        mock_conn = MagicMock(spec=Connection)
        mock_conn.send = AsyncMock()
        mock_conn.close = AsyncMock()

        async with MyAgent("test", storage=MemoryStorage()) as agent:
            agent._connections.add(mock_conn)
            agent.enqueue("close", (1000, "Normal"), connection=mock_conn)
            await agent._mailbox.join()

        assert (1000, "Normal") in closed


class TestAgentEmit:
    """Tests for Agent event emission."""

    @pytest.mark.asyncio
    async def test_emit_returns_event_id(self):
        """emit() should return incrementing event IDs."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        id1 = await agent.emit("status", {"text": "Hello"})
        id2 = await agent.emit("status", {"text": "World"})

        assert id1 == 1
        assert id2 == 2
        await agent.stop()

    @pytest.mark.asyncio
    async def test_emit_logs_to_event_log(self):
        """emit() should add events to _event_log."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        await agent.emit("status", {"text": "Hello"})
        await agent.emit("chunk", {"content": "World"})

        assert len(agent._event_log) == 2
        assert agent._event_log[0]["type"] == "status"
        assert agent._event_log[1]["type"] == "chunk"
        await agent.stop()

    @pytest.mark.asyncio
    async def test_emit_persists_when_enabled(self):
        """emit() should persist to storage when enable_event_log=True."""
        storage = MemoryStorage()

        class DurableAgent(Agent):
            enable_event_log = True

        agent = DurableAgent("test", storage=storage)
        await agent.start()

        await agent.emit("status", {"text": "Hello"})

        stored = await storage.get("test:event_log")
        assert stored is not None
        assert len(stored) == 1
        await agent.stop()

    @pytest.mark.asyncio
    async def test_emit_does_not_persist_when_disabled(self):
        """emit() should not persist when enable_event_log=False."""
        storage = MemoryStorage()
        agent = Agent("test", storage=storage)
        await agent.start()

        await agent.emit("status", {"text": "Hello"})

        stored = await storage.get("test:event_log")
        assert stored is None
        await agent.stop()

    @pytest.mark.asyncio
    async def test_emit_calls_handler(self):
        """emit() should call _emit_handler if set."""
        received = []

        async def handler(event, data, event_id):
            received.append((event, data, event_id))

        agent = Agent("test", storage=MemoryStorage())
        agent._emit_handler = handler
        await agent.start()

        await agent.emit("status", {"text": "Hello"})

        assert len(received) == 1
        assert received[0] == ("status", {"text": "Hello"}, 1)
        await agent.stop()

    @pytest.mark.asyncio
    async def test_emit_works_without_handler(self):
        """emit() should work even without _emit_handler."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        # Should not raise
        event_id = await agent.emit("status", {"text": "Hello"})
        assert event_id == 1
        await agent.stop()


class TestAgentReplayEvents:
    """Tests for event replay."""

    @pytest.mark.asyncio
    async def test_replay_events_yields_events(self):
        """replay_events() should yield Event objects."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        await agent.emit("a", {"n": 1})
        await agent.emit("b", {"n": 2})
        await agent.emit("c", {"n": 3})

        events = []
        async for event in agent.replay_events(after_id=0):
            events.append(event)

        assert len(events) == 3
        assert events[0].type == "a"
        assert events[1].type == "b"
        assert events[2].type == "c"
        await agent.stop()

    @pytest.mark.asyncio
    async def test_replay_events_filters_by_id(self):
        """replay_events() should only yield events after the given ID."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        await agent.emit("a", {"n": 1})
        await agent.emit("b", {"n": 2})
        await agent.emit("c", {"n": 3})

        events = []
        async for event in agent.replay_events(after_id=1):
            events.append(event)

        assert len(events) == 2
        assert events[0].id == 2
        assert events[1].id == 3
        await agent.stop()

    @pytest.mark.asyncio
    async def test_replay_events_empty_when_no_events(self):
        """replay_events() should yield nothing when log is empty."""
        agent = Agent("test", storage=MemoryStorage())
        await agent.start()

        events = []
        async for event in agent.replay_events(after_id=0):
            events.append(event)

        assert events == []
        await agent.stop()


class TestAgentEventLogPersistence:
    """Tests for event log persistence."""

    @pytest.mark.asyncio
    async def test_event_log_loaded_on_start(self):
        """Event log should be loaded from storage on start."""
        storage = MemoryStorage()
        await storage.set(
            "test:event_log",
            [
                {"id": 1, "type": "a", "data": {}},
                {"id": 2, "type": "b", "data": {}},
            ],
        )

        class DurableAgent(Agent):
            enable_event_log = True

        agent = DurableAgent("test", storage=storage)
        await agent.start()

        assert len(agent._event_log) == 2
        assert agent._event_counter == 2
        await agent.stop()

    @pytest.mark.asyncio
    async def test_event_counter_continues_from_log(self):
        """Event counter should continue from max ID in log."""
        storage = MemoryStorage()
        await storage.set(
            "test:event_log",
            [
                {"id": 5, "type": "a", "data": {}},
                {"id": 10, "type": "b", "data": {}},
            ],
        )

        class DurableAgent(Agent):
            enable_event_log = True

        agent = DurableAgent("test", storage=storage)
        await agent.start()

        event_id = await agent.emit("c", {})
        assert event_id == 11
        await agent.stop()

    @pytest.mark.asyncio
    async def test_clear_event_log(self):
        """clear_event_log() should reset log and counter."""
        storage = MemoryStorage()

        class DurableAgent(Agent):
            enable_event_log = True

        agent = DurableAgent("test", storage=storage)
        await agent.start()

        await agent.emit("a", {})
        await agent.emit("b", {})
        await agent.clear_event_log()

        assert len(agent._event_log) == 0
        assert agent._event_counter == 0
        await agent.stop()



"""
Example Cloudflare Worker script using ai-query Durable Objects.

This example demonstrates a complete setup for a stateful "Researcher" agent
running on Cloudflare Workers.

Prerequisites:
- A Cloudflare account
- Wrangler CLI installed (`npm install -g wrangler`)

Setup:
1. Create a `wrangler.toml` (see below)
2. Run `wrangler deploy`

"""

# --- 1. The Agent Logic (Platform Agnostic) ---
from ai_query import Agent, action

class ResearcherAgent(Agent):
    """An agent that performs research and remembers it."""
    
    @action
    async def research(self, topic: str):
        # Emit status update (sent via WebSocket if connected)
        await self.emit("status", {"text": f"Starting research on {topic}..."})
        
        # Persist state automatically using Durable Object storage
        history = self.state.get("history", [])
        history.append(topic)
        self.state["history"] = history
        
        # Simulate work
        import asyncio
        await asyncio.sleep(1) 
        
        result = f"Detailed report on {topic}"
        
        await self.emit("result", {"summary": result})
        return result

    @action
    async def get_history(self):
        return self.state.get("history", [])

# --- 2. The Cloudflare Adapter ---
from ai_query.adapters.cloudflare import AgentDO, CloudflareRegistry

class ResearcherDO(AgentDO):
    """The Durable Object wrapper.
    
    This class bridges the Cloudflare DO lifecycle with the Agent lifecycle.
    It inherits from `workers.DurableObject` and handles storage/networking.
    """
    agent_class = ResearcherAgent


# --- 3. The Worker Entry Point ---
async def fetch(request, env):
    """Main Worker fetch handler.
    
    Routes incoming HTTP and WebSocket requests to the correct Durable Object.
    """
    
    # The Registry knows how to find DOs based on bindings
    registry = CloudflareRegistry(env)
    
    # Register the route.
    # Pattern: "researcher-.*" matches agent IDs like "researcher-123"
    # Binding: env.RESEARCHER (defined in wrangler.toml)
    registry.register("researcher-.*", env.RESEARCHER)
    
    # Handles:
    # - POST /agent/researcher-123/chat (Chat)
    # - POST /agent/researcher-123/invoke (Action)
    # - WebSocket /agent/researcher-123 (Real-time)
    return await registry.handle_request(request)


# --- Appendix: Example wrangler.toml ---
"""
name = "ai-query-worker"
main = "worker.py"
compatibility_date = "2024-04-03"

[durable_objects]
bindings = [
  { name = "RESEARCHER", class_name = "ResearcherDO" }
]

[[migrations]]
tag = "v1"
new_classes = ["ResearcherDO"]

[observability]
enabled = true
"""



"""Real-time Task Assistant with AI.

Usage:
    uv run examples/realtime/server.py

Then connect with the client:
    uv run examples/realtime/client.py --user alice

Or use curl:
    curl http://localhost:8080/events  # SSE stream
    curl -X POST http://localhost:8080/chat -H "Content-Type: application/json" -d '{"message": "Add a task: buy groceries"}'
"""

from ai_query import tool, Field
from ai_query.agents import Agent, MemoryStorage
from ai_query.providers import google


class TaskAssistant(Agent):
    """Real-time task assistant with AI support."""

    enable_event_log = True  # Enable event replay on reconnect

    def __init__(self):
        @tool(description="Add a new task to the list")
        async def add_task(task: str = Field(description="Task to add")) -> str:
            tasks = self.state.get("tasks", []) + [task]
            await self.update_state(tasks=tasks)
            await self.emit("task_added", {"task": task, "count": len(tasks)})
            return f"Added task #{len(tasks)}: {task}"

        @tool(description="Mark a task as complete")
        async def complete_task(task_number: int = Field(description="Task number to complete")) -> str:
            tasks = self.state.get("tasks", [])
            if task_number < 1 or task_number > len(tasks):
                return f"Invalid task number. You have {len(tasks)} tasks."
            task = tasks[task_number - 1]
            remaining = [t for i, t in enumerate(tasks) if i != task_number - 1]
            completed = self.state.get("completed", []) + [task]
            await self.update_state(tasks=remaining, completed=completed)
            await self.emit("task_completed", {"task": task, "remaining": len(remaining)})
            return f"Completed: {task}"

        @tool(description="List all current tasks")
        async def list_tasks() -> str:
            tasks = self.state.get("tasks", [])
            if not tasks:
                return "No pending tasks."
            return "Tasks:\n" + "\n".join([f"{i+1}. {t}" for i, t in enumerate(tasks)])

        super().__init__(
            "task-assistant",
            model=google("gemini-2.0-flash"),
            system="""You are a task assistant. Help users plan and complete tasks.
            Use the available tools to track tasks and progress.""",
            storage=MemoryStorage(),
            initial_state={"tasks": [], "completed": []},
            tools={"add_task": add_task, "complete_task": complete_task, "list_tasks": list_tasks},
        )

    async def on_start(self):
        task_count = len(self.state.get("tasks", []))
        print(f"Task Assistant started. Pending tasks: {task_count}")

    async def on_connect(self, connection, ctx):
        await super().on_connect(connection, ctx)
        task_count = len(self.state.get("tasks", []))
        if task_count > 0:
            await connection.send(f"Welcome back! You have {task_count} pending tasks.")
        else:
            await connection.send("Hello! I'm your task assistant. How can I help?")

    async def on_message(self, connection, message):
        """Handle WebSocket messages with streaming."""
        await self.emit("chat_start", {"message": message})

        async for chunk in self.stream(message):
            await self.emit("chunk", {"content": chunk})
            await connection.send(chunk)

        await self.emit("chat_complete", {})


if __name__ == "__main__":
    print("Task Assistant Server")
    print("=" * 40)
    print("Endpoints:")
    print("  POST /chat   - Send message")
    print("  GET  /events - SSE stream")
    print("  WS   /ws     - WebSocket")
    print("  GET  /state  - Agent state")
    print()
    TaskAssistant().serve(host="localhost", port=8080)



"""In-memory storage implementation."""

from __future__ import annotations

from typing import Any


class MemoryStorage:
    """In-memory storage for development and testing.

    Data is lost when the process exits.

    Example:
        agent = Agent("assistant", storage=MemoryStorage())
    """

    def __init__(self) -> None:
        self._data: dict[str, Any] = {}

    async def get(self, key: str) -> Any | None:
        return self._data.get(key)

    async def set(self, key: str, value: Any) -> None:
        self._data[key] = value

    async def delete(self, key: str) -> None:
        self._data.pop(key, None)

    async def keys(self, prefix: str = "") -> list[str]:
        if not prefix:
            return list(self._data.keys())
        return [k for k in self._data.keys() if k.startswith(prefix)]

    def clear(self) -> None:
        """Clear all data."""
        self._data.clear()



import asyncio
import json
import sys
from unittest.mock import MagicMock, AsyncMock

# Mock 'js' module before importing adapters
sys.modules["js"] = MagicMock()
sys.modules["pyodide"] = MagicMock()
sys.modules["pyodide.ffi"] = MagicMock()
sys.modules["aiohttp"] = MagicMock()
sys.modules["aiohttp.web"] = MagicMock()
sys.modules["httpx"] = MagicMock()
sys.modules["fastapi"] = MagicMock()

# Setup JS Mocks
js = sys.modules["js"]
js.Response.new = MagicMock(
    side_effect=lambda body, init=None: {"body": body, "init": init}
)
js.Object.fromEntries = MagicMock(side_effect=lambda x: x)


# Mock Request
class MockRequest:
    def __init__(self, method, body=None, url="http://worker/"):
        self.method = method
        self._body = body
        self.headers = {}
        self.url = url

    async def json(self):
        return self._body


# Import logic to test
from ai_query import Agent
from ai_query.adapters.cloudflare import AgentDO, CloudflareRegistry


# Define Agent
class TestAgent(Agent):
    async def handle_request(self, request):
        return {"result": "ok", "echo": request}


# Define DO
class TestDO(AgentDO):
    agent_class = TestAgent


async def test_do_fetch():
    print("Testing AgentDO.fetch()...")

    # Mock Ctx & Env
    ctx = MagicMock()
    ctx.id = "test-id"
    ctx.storage = AsyncMock()  # Mock Storage

    env = MagicMock()

    # Instantiate DO
    do = TestDO(ctx, env)

    # Check Storage Init
    assert do.storage._storage == ctx.storage
    print("✓ Storage initialized")

    # Simulate POST Request
    req_body = {"action": "invoke", "payload": {"foo": "bar"}}
    req = MockRequest("POST", body=req_body)

    # Force agent start (async)
    # The actual 'fetch' calls 'agent.start()' inside.
    # Agent.start calls storage.get ...
    ctx.storage.get.return_value = None  # No existing state

    response = await do.fetch(req)

    # Verify Response
    body_str = response["body"]
    body_json = json.loads(body_str)
    print(f"DEBUG Response: {body_json}")

    # The Agent logic for 'invoke' -> 'action' -> or handle_invoke
    # Wait, TestAgent overrides handle_request!
    # So it should return {"result": "ok", ...}

    assert body_json["result"] == "ok"
    print("✓ Fetch POST handled correctly")


async def test_registry():
    print("Testing CloudflareRegistry...")
    env = MagicMock()
    registry = CloudflareRegistry(env)

    # Mock Binding
    binding = MagicMock()
    stub = MagicMock()
    stub.fetch = AsyncMock(return_value="stub_response")
    # binding.getByName returns the stub directly
    binding.getByName.return_value = stub

    registry.register("test-.*", binding)

    # Test Match
    req = MockRequest("GET", url="http://worker/agent/test-123/chat")
    # Need to mock js.URL inside registry
    js.URL.new = lambda url: MagicMock(pathname="/agent/test-123/chat")

    await registry.handle_request(req)

    binding.getByName.assert_called_with("test-123")
    stub.fetch.assert_called_with(req)
    print("✓ Registry routing working")


async def test_websocket_hibernation():
    print("Testing WebSocket Hibernation...")
    ctx = MagicMock()
    ctx.id = "test-ws"
    ctx.storage = AsyncMock()
    ctx.storage.get.return_value = None
    env = MagicMock()

    do = TestDO(ctx, env)

    # Mock WebSocket
    ws = MagicMock()

    # Simulate webSocketMessage
    # It should start the agent if not running
    await do.webSocketMessage(ws, "hello")

    if do.agent._state is None:
        # It tried to start
        pass

    # Check if agent.on_message was called?
    # We can't easily check internal agent.on_message without mocking agent instance or its connection.
    # But we can check if it didn't crash.
    print("✓ webSocketMessage handled")

    # Simulate webSocketClose
    await do.webSocketClose(ws, 1000, "normal", True)
    print("✓ webSocketClose handled")


async def test_serialization():
    print("Testing Robust Serialization...")
    from ai_query.agents.storage.cloudflare import DurableObjectStorage

    # Mock storage
    mock_storage = AsyncMock()
    # Mock put/get
    store = {}

    async def put(k, v):
        store[k] = v

    async def get(k):
        return store.get(k)

    mock_storage.put.side_effect = put
    mock_storage.get.side_effect = get

    do_storage = DurableObjectStorage(mock_storage)

    # Test Data with Set and Tuple
    complex_data = {"tags": {"a", "b", "c"}, "coords": (10, 20), "normal": "string"}

    # Save
    await do_storage.set("complex", complex_data)

    # Verify it was serialized (check raw store)
    raw = store["complex"]
    assert "__ai_query_type__" in raw
    print("✓ Data serialized with custom types")

    # Load
    loaded = await do_storage.get("complex")

    print(f"DEBUG: loaded type: {type(loaded)}")
    print(f"DEBUG: loaded data: {loaded}")
    print(f"DEBUG: coords type: {type(loaded.get('coords'))}")

    assert isinstance(loaded["tags"], set)
    assert loaded["tags"] == {"a", "b", "c"}
    # JSON converts tuples to lists by default; Encoder.default isn't called for tuples.
    # We accept this degradation as standard for JSON storage.
    assert isinstance(loaded["coords"], list)
    assert loaded["coords"] == [10, 20]
    print("✓ Data deserialized correctly (Set restored, Tuple -> List)")


def main():
    loop = asyncio.new_event_loop()
    loop.run_until_complete(test_do_fetch())
    loop.run_until_complete(test_registry())
    loop.run_until_complete(test_websocket_hibernation())
    loop.run_until_complete(test_serialization())
    print("All verification tests passed!")


if __name__ == "__main__":
    main()



"""Unit Converter Agent

Demonstrates:
- Simple synchronous tool execution
- Mathematical conversions
- Clean tool definitions
"""

import asyncio
from ai_query import generate_text, tool, Field, step_count_is
from ai_query.providers import google


# --- Conversion Tools ---

@tool(description="Convert temperature between Celsius, Fahrenheit, and Kelvin.")
def convert_temperature(
    value: float = Field(description="The temperature value"),
    from_unit: str = Field(description="Source unit (celsius/fahrenheit/kelvin)"),
    to_unit: str = Field(description="Target unit (celsius/fahrenheit/kelvin)")
) -> str:
    """Convert temperature between Celsius, Fahrenheit, and Kelvin."""
    print(f"  [Convert] {value} {from_unit} -> {to_unit}")

    # Normalize units
    from_unit = from_unit.lower()[0]  # c, f, or k
    to_unit = to_unit.lower()[0]

    # Convert to Celsius first
    if from_unit == "f":
        celsius = (value - 32) * 5 / 9
    elif from_unit == "k":
        celsius = value - 273.15
    else:
        celsius = value

    # Convert from Celsius to target
    if to_unit == "f":
        result = celsius * 9 / 5 + 32
        unit_name = "Fahrenheit"
    elif to_unit == "k":
        result = celsius + 273.15
        unit_name = "Kelvin"
    else:
        result = celsius
        unit_name = "Celsius"

    return f"{value} {from_unit.upper()} = {result:.2f} {unit_name}"


@tool(description="Convert length between mm, cm, m, km, inches, feet, yards, miles.")
def convert_length(
    value: float = Field(description="The length value"),
    from_unit: str = Field(description="Source unit"),
    to_unit: str = Field(description="Target unit")
) -> str:
    """Convert length between various units."""
    print(f"  [Convert] {value} {from_unit} -> {to_unit}")

    # Conversion factors to meters
    to_meters = {
        "mm": 0.001, "cm": 0.01, "m": 1, "km": 1000,
        "in": 0.0254, "ft": 0.3048, "yd": 0.9144, "mi": 1609.34,
        "inch": 0.0254, "inches": 0.0254, "foot": 0.3048, "feet": 0.3048,
        "yard": 0.9144, "yards": 0.9144, "mile": 1609.34, "miles": 1609.34,
        "meter": 1, "meters": 1, "kilometer": 1000, "kilometers": 1000,
    }

    from_unit = from_unit.lower()
    to_unit = to_unit.lower()

    if from_unit not in to_meters or to_unit not in to_meters:
        return f"Unknown unit. Supported: {', '.join(set(to_meters.keys()))}"

    # Convert to meters, then to target
    meters = value * to_meters[from_unit]
    result = meters / to_meters[to_unit]

    return f"{value} {from_unit} = {result:.4f} {to_unit}"


@tool(description="Convert weight between mg, g, kg, ounces, pounds.")
def convert_weight(
    value: float = Field(description="The weight value"),
    from_unit: str = Field(description="Source unit"),
    to_unit: str = Field(description="Target unit")
) -> str:
    """Convert weight/mass between various units."""
    print(f"  [Convert] {value} {from_unit} -> {to_unit}")

    # Conversion factors to grams
    to_grams = {
        "mg": 0.001, "g": 1, "kg": 1000,
        "oz": 28.3495, "lb": 453.592, "lbs": 453.592,
        "ounce": 28.3495, "ounces": 28.3495,
        "pound": 453.592, "pounds": 453.592,
        "gram": 1, "grams": 1, "kilogram": 1000, "kilograms": 1000,
    }

    from_unit = from_unit.lower()
    to_unit = to_unit.lower()

    if from_unit not in to_grams or to_unit not in to_grams:
        return f"Unknown unit. Supported: {', '.join(set(to_grams.keys()))}"

    grams = value * to_grams[from_unit]
    result = grams / to_grams[to_unit]

    return f"{value} {from_unit} = {result:.4f} {to_unit}"


@tool(description="Convert currency between major world currencies (USD, EUR, GBP, JPY, etc.). Uses approximate rates.")
def convert_currency(
    amount: float = Field(description="The amount to convert"),
    from_currency: str = Field(description="Source currency code (e.g., USD)"),
    to_currency: str = Field(description="Target currency code (e.g., EUR)")
) -> str:
    """Convert currency using approximate rates (for demo purposes)."""
    print(f"  [Convert] {amount} {from_currency} -> {to_currency}")

    # Approximate rates to USD (demo only - use real API for production)
    to_usd = {
        "USD": 1, "EUR": 1.08, "GBP": 1.27, "JPY": 0.0067,
        "CAD": 0.74, "AUD": 0.65, "CHF": 1.13, "CNY": 0.14,
        "INR": 0.012, "MXN": 0.058, "BRL": 0.20, "KRW": 0.00075,
    }

    from_currency = from_currency.upper()
    to_currency = to_currency.upper()

    if from_currency not in to_usd or to_currency not in to_usd:
        return f"Unknown currency. Supported: {', '.join(to_usd.keys())}"

    usd = amount * to_usd[from_currency]
    result = usd / to_usd[to_currency]

    return f"{amount} {from_currency} ≈ {result:.2f} {to_currency} (approximate rate)"


# --- Main ---

async def main():
    print("Unit Converter Agent")
    print("=" * 50)

    questions = [
        "What is 100 degrees Fahrenheit in Celsius?",
        "Convert 5 miles to kilometers",
        "How many pounds is 75 kilograms?",
        "I have 1000 USD, how much is that in EUR, GBP, and JPY?",
        "If water boils at 100°C and freezes at 0°C, what are those in Fahrenheit?",
    ]

    tools = {
        "convert_temperature": convert_temperature,
        "convert_length": convert_length,
        "convert_weight": convert_weight,
        "convert_currency": convert_currency,
    }

    for q in questions:
        print(f"\nQ: {q}")

        result = await generate_text(
            model=google("gemini-2.0-flash"),
            system="You are a helpful unit converter. Use the conversion tools to answer questions accurately.",
            prompt=q,
            tools=tools,
            stop_when=step_count_is(4),
        )

        print(f"A: {result.text}")
        print(f"   [{result.usage.total_tokens} tokens]")


if __name__ == "__main__":
    asyncio.run(main())



"""Example: Using Agent with stream_text directly.

This shows how to use stream_text directly within an Agent for full control
over AI interactions, while using emit() for events.

Usage:
    uv run examples/agent_direct.py

Connect:
    curl http://localhost:8080/events  # SSE stream
    curl -X POST http://localhost:8080/chat -H "Content-Type: application/json" -d '{"message": "Hello @ai"}'
"""

from ai_query import stream_text, tool, Field
from ai_query.agents import Agent, MemoryStorage
from ai_query.providers import google


class CustomAIAgent(Agent):
    """Agent using stream_text directly instead of agent.stream()."""

    def __init__(self):
        super().__init__(
            "custom-agent",
            storage=MemoryStorage(),
            initial_state={"query_count": 0, "last_topic": None},
        )

    async def on_connect(self, connection, ctx):
        await super().on_connect(connection, ctx)
        username = ctx.metadata.get("username", "Anonymous")
        connection.username = username
        await connection.send(f"Welcome {username}! Ask me anything with @ai")

    async def on_message(self, connection, message):
        username = getattr(connection, "username", "Anonymous")

        if "@ai" in message.lower():
            await self._handle_ai_query(connection, username, message)
        else:
            await connection.send(f"Echo: {message}")

    async def _handle_ai_query(self, connection, username: str, message: str):
        """Handle AI query using stream_text directly."""

        # Update state
        await self.update_state(
            query_count=self.state.get("query_count", 0) + 1,
            last_topic=message[:50],
        )

        # Define tools with access to agent via closure
        @tool(description="Save a note for the user")
        async def save_note(note: str = Field(description="Note to save")) -> str:
            notes = self.state.get("notes", []) + [note]
            await self.update_state(notes=notes)
            await self.emit("note_saved", {"note": note})
            return f"Saved: {note}"

        @tool(description="Get the current query count")
        def get_stats() -> str:
            return f"Total queries: {self.state.get('query_count', 0)}"

        # Emit start event
        await self.emit("ai_start", {"username": username, "query": message})

        # Use stream_text directly for full control
        result = stream_text(
            model=google("gemini-2.0-flash"),
            system=f"""You are a helpful AI assistant.
            You're talking to {username}.
            Be concise and helpful.
            Use tools when appropriate.""",
            prompt=message,
            tools={"save_note": save_note, "get_stats": get_stats},
        )

        # Stream chunks via emit
        full_response = ""
        async for chunk in result.text_stream:
            full_response += chunk
            await self.emit("chunk", {"content": chunk})

        # Emit completion
        await self.emit("ai_complete", {"response": full_response})

        # Send to WebSocket client
        await connection.send(full_response)

        print(f"AI responded to {username} (query #{self.state.get('query_count', 0)})")


if __name__ == "__main__":
    print("Custom AI Agent Server")
    print("=" * 40)
    print("This example shows using Agent + stream_text directly")
    print("for full control over AI interactions.")
    print()
    print("Endpoints:")
    print("  POST /chat   - Send message")
    print("  GET  /events - SSE stream")
    print("  WS   /ws     - WebSocket")
    print("  GET  /state  - Agent state")
    print()
    CustomAIAgent().serve(port=8080)



"""MCP utility functions."""

from __future__ import annotations

from ai_query.types import ToolSet
from ai_query.mcp.types import MCPServer


def merge_tools(*tool_sources: ToolSet | MCPServer) -> ToolSet:
    """Merge multiple tool sources into a single ToolSet.

    This is useful when you want to combine tools from multiple MCP servers
    or mix MCP tools with locally defined tools.

    Args:
        *tool_sources: ToolSet dicts or MCPServer instances to merge.

    Returns:
        A merged ToolSet containing all tools.

    Example:
        >>> local_tools = {"calculator": calc_tool}
        >>> async with mcp("python", "weather_server.py") as weather:
        ...     async with mcp("python", "search_server.py") as search:
        ...         all_tools = merge_tools(local_tools, weather, search)
        ...         result = await generate_text(
        ...             model=openai("gpt-4o"),
        ...             prompt="Search and calculate",
        ...             tools=all_tools,
        ...         )
    """
    merged: ToolSet = {}
    for source in tool_sources:
        if isinstance(source, MCPServer):
            merged.update(source.tools)
        else:
            merged.update(source)
    return merged



"""Core types for ai-query library."""

from __future__ import annotations

import asyncio
import inspect
from dataclasses import dataclass, field
from typing import (
    Any,
    Literal,
    Union,
    AsyncIterator,
    Callable,
    Awaitable,
    TYPE_CHECKING,
    get_type_hints,
    get_origin,
    get_args,
    overload,
)

# Message types
Role = Literal["system", "user", "assistant", "tool"]


# =============================================================================
# Field - Pydantic-style parameter metadata
# =============================================================================

# Sentinel for required fields (like Pydantic's ...)
_MISSING = object()


class Field:
    """Define metadata for a tool parameter."""

    def __init__(
        self,
        description: str = "",
        default: Any = _MISSING,
        enum: Union[list[Any], None] = None,
        min_value: Union[float, None] = None,
        max_value: Union[float, None] = None,
    ) -> None:
        self.description = description
        self.default = default
        self.enum = enum
        self.min_value = min_value
        self.max_value = max_value

    def __repr__(self) -> str:
        return f"Field(description={self.description!r}, default={self.default!r})"


def _python_type_to_json_schema(py_type: Any) -> dict[str, Any]:
    """Convert a Python type hint to JSON Schema."""
    origin = get_origin(py_type)
    args = get_args(py_type)

    if origin is Union:
        non_none = [a for a in args if a is not type(None)]
        if len(non_none) == 1:
            return _python_type_to_json_schema(non_none[0])
        return {"anyOf": [_python_type_to_json_schema(t) for t in non_none]}

    if origin is list:
        item_schema = _python_type_to_json_schema(args[0]) if args else {}
        return {"type": "array", "items": item_schema}

    if origin is dict:
        return {"type": "object"}

    if origin is Literal:
        return {"type": "string", "enum": list(args)}

    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
    }

    if py_type in type_map:
        return {"type": type_map[py_type]}

    return {"type": "string"}


# =============================================================================
# Tool Types
# =============================================================================


@dataclass
class Tool:
    description: str
    parameters: dict[str, Any]
    execute: Union[Callable[..., Any], Callable[..., Awaitable[Any]]]

    async def run(self, **kwargs: Any) -> Any:
        sig = inspect.signature(self.execute)
        bound_args = sig.bind_partial(**kwargs)

        for name, param in sig.parameters.items():
            if name not in bound_args.arguments and isinstance(param.default, Field):
                if param.default.default is not _MISSING:
                    kwargs[name] = param.default.default
        
        result = self.execute(**kwargs)
        if inspect.isawaitable(result):
            return await result
        return result


@overload
def tool(func: Callable[..., Any]) -> Tool: ...

@overload
def tool(*, description: Union[str, None] = None) -> Callable[[Callable[..., Any]], Tool]: ...

@overload
def tool(
    *,
    description: str,
    parameters: dict[str, Any],
    execute: Union[Callable[..., Any], Callable[..., Awaitable[Any]]],
) -> Tool: ...


def tool(
    func: Union[Callable[..., Any], None] = None,
    *,
    description: Union[str, None] = None,
    parameters: Union[dict[str, Any], None] = None,
    execute: Union[Callable[..., Any], Callable[..., Awaitable[Any]], None] = None,
) -> Union[Tool, Callable[[Callable[..., Any]], Tool]]:
    if execute is not None and parameters is not None and description is not None:
        return Tool(description=description, parameters=parameters, execute=execute)

    def _create_tool_from_function(fn: Callable[..., Any]) -> Tool:
        tool_description = description or ""
        if not tool_description and fn.__doc__:
            tool_description = fn.__doc__.strip().split("\n")[0].strip()
        if not tool_description:
            tool_description = f"Execute the {fn.__name__} function"

        try:
            hints = get_type_hints(fn)
        except Exception:
            hints = {}

        sig = inspect.signature(fn)
        properties: dict[str, Any] = {}
        required: list[str] = []

        for param_name, param in sig.parameters.items():
            if param_name in ("self", "cls", "return"):
                continue

            param_type = hints.get(param_name, str)
            prop = _python_type_to_json_schema(param_type)

            if isinstance(param.default, Field):
                field_meta = param.default
                if field_meta.description:
                    prop["description"] = field_meta.description
                if field_meta.enum:
                    prop["enum"] = field_meta.enum
                if field_meta.min_value is not None:
                    prop["minimum"] = field_meta.min_value
                if field_meta.max_value is not None:
                    prop["maximum"] = field_meta.max_value
                if field_meta.default is _MISSING:
                    required.append(param_name)
            elif param.default is inspect.Parameter.empty:
                required.append(param_name)
            
            properties[param_name] = prop

        schema = {"type": "object", "properties": properties, "required": required}
        return Tool(description=tool_description, parameters=schema, execute=fn)

    if func is not None:
        return _create_tool_from_function(func)
    else:
        return _create_tool_from_function

ToolSet = dict[str, Tool]


@dataclass
class ToolCall:
    id: str
    name: str
    arguments: dict[str, Any]
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class ToolResult:
    tool_call_id: str
    tool_name: str
    result: Any
    is_error: bool = False


@dataclass
class ToolCallPart:
    type: Literal["tool_call"] = "tool_call"
    tool_call: Union[ToolCall, None] = None


@dataclass
class ToolResultPart:
    type: Literal["tool_result"] = "tool_result"
    tool_result: Union[ToolResult, None] = None


# =============================================================================
# Stop Conditions
# =============================================================================


@dataclass
class StepResult:
    text: str
    tool_calls: list[ToolCall]
    tool_results: list[ToolResult]
    finish_reason: Union[str, None] = None


StopCondition = Callable[[list[StepResult]], Union[bool, Awaitable[bool]]]


def step_count_is(count: int) -> StopCondition:
    def condition(steps: list[StepResult]) -> bool:
        return len(steps) >= count
    return condition


def has_tool_call(tool_name: str) -> StopCondition:
    def condition(steps: list[StepResult]) -> bool:
        if not steps:
            return False
        last_step = steps[-1]
        return any(tc.name == tool_name for tc in last_step.tool_calls)
    return condition


# =============================================================================
# Step Callbacks
# =============================================================================


@dataclass
class StepStartEvent:
    step_number: int
    messages: list["Message"]
    tools: Union[ToolSet, None]


@dataclass
class StepFinishEvent:
    step_number: int
    step: StepResult
    text: str
    usage: Usage
    steps: list[StepResult]


OnStepStart = Callable[[StepStartEvent], Union[None, Awaitable[None]]]
OnStepFinish = Callable[[StepFinishEvent], Union[None, Awaitable[None]]]


@dataclass
class TextPart:
    type: Literal["text"] = "text"
    text: str = ""


@dataclass
class ImagePart:
    type: Literal["image"] = "image"
    image: Union[str, bytes] = b""
    media_type: Union[str, None] = None


@dataclass
class FilePart:
    type: Literal["file"] = "file"
    data: Union[str, bytes] = b""
    media_type: str = ""


ContentPart = Union[TextPart, ImagePart, FilePart]


@dataclass
class Message:
    role: Role
    content: Union[str, list[ContentPart]]

    def to_dict(self) -> dict[str, Any]:
        if isinstance(self.content, str):
            return {"role": self.role, "content": self.content}
        
        content_list = []
        for p in self.content:
            if isinstance(p, dict):
                content_list.append(p)
            else:
                part_dict: dict[str, Any] = {"type": p.type}
                if isinstance(p, TextPart):
                    part_dict["text"] = p.text
                elif isinstance(p, ImagePart):
                    part_dict["image"] = p.image
                    if p.media_type:
                        part_dict["media_type"] = p.media_type
                elif isinstance(p, FilePart):
                    part_dict["data"] = p.data
                    part_dict["media_type"] = p.media_type
                content_list.append(part_dict)
        
        return {"role": self.role, "content": content_list}


@dataclass
class Usage:
    input_tokens: int = 0
    output_tokens: int = 0
    cached_tokens: int = 0
    total_tokens: int = 0


@dataclass
class GenerateTextResult:
    text: str
    steps: list[StepResult] = field(default_factory=list)
    finish_reason: Union[str, None] = None
    usage: Union[Usage, None] = None
    response: dict[str, Any] = field(default_factory=dict)
    provider_metadata: dict[str, Any] = field(default_factory=dict)

    @property
    def tool_calls(self) -> list[ToolCall]:
        calls = []
        for step in self.steps:
            calls.extend(step.tool_calls)
        return calls

    @property
    def tool_results(self) -> list[ToolResult]:
        results = []
        for step in self.steps:
            results.extend(step.tool_results)
        return results


@dataclass
class AgentEvent:
    id: int
    type: str
    data: dict[str, Any]
    created_at: float


@dataclass
class StreamTextResult:
    text: str
    finish_reason: Union[str, None] = None
    usage: Union[Usage, None] = None


@dataclass
class StreamChunk:
    text: str = ""
    is_final: bool = False
    usage: Union[Usage, None] = None
    finish_reason: Union[str, None] = None
    tool_calls: Union[list[ToolCall], None] = None


class TextStreamResult:
    def __init__(self, stream: AsyncIterator[StreamChunk], steps: Union[list[StepResult], None] = None) -> None:
        self._stream = stream
        self._chunks: list[str] = []
        self._usage: Union[Usage, None] = None
        self._finish_reason: Union[str, None] = None
        self._steps: list[StepResult] = steps or []
        self._done = False
        self._done_event = asyncio.Event()
        self._consumed = False

    async def _consume_stream(self) -> AsyncIterator[str]:
        if self._consumed:
            for chunk in self._chunks:
                yield chunk
            return

        self._consumed = True
        async for chunk in self._stream:
            if chunk.is_final:
                self._usage = chunk.usage
                self._finish_reason = chunk.finish_reason
            elif chunk.text:
                self._chunks.append(chunk.text)
                yield chunk.text

        self._done = True
        self._done_event.set()

    @property
    def text_stream(self) -> AsyncIterator[str]:
        return self._consume_stream()

    async def _wait_for_completion(self) -> None:
        if not self._done:
            if not self._consumed:
                async for _ in self._consume_stream():
                    pass
            else:
                await self._done_event.wait()

    @property
    async def text(self) -> str:
        await self._wait_for_completion()
        return "".join(self._chunks)

    @property
    async def usage(self) -> Union[Usage, None]:
        await self._wait_for_completion()
        return self._usage

    @property
    async def finish_reason(self) -> Union[str, None]:
        await self._wait_for_completion()
        return self._finish_reason

    @property
    async def steps(self) -> list[StepResult]:
        await self._wait_for_completion()
        return self._steps

    def __aiter__(self) -> AsyncIterator[str]:
        return self.text_stream

ProviderOptions = dict[str, dict[str, Any]]


# =============================================================================
# Embedding Types
# =============================================================================


@dataclass
class EmbeddingUsage:
    tokens: int = 0


@dataclass
class EmbedResult:
    value: str
    embedding: list[float]
    usage: EmbeddingUsage
    provider_metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class EmbedManyResult:
    values: list[str]
    embeddings: list[list[float]]
    usage: EmbeddingUsage
    provider_metadata: dict[str, Any] = field(default_factory=dict)


# =============================================================================
# Abort / Cancellation
# =============================================================================


class AbortError(Exception):
    def __init__(self, reason: Union[str, None] = None):
        self.reason = reason
        super().__init__(reason or "Operation aborted")


class AbortSignal:
    def __init__(self) -> None:
        self._aborted = False
        self._reason: Union[str, None] = None
        self._listeners: list[Callable[[], None]] = []
        self._event = asyncio.Event()

    @property
    def aborted(self) -> bool:
        return self._aborted

    @property
    def reason(self) -> Union[str, None]:
        return self._reason

    def throw_if_aborted(self) -> None:
        if self._aborted:
            raise AbortError(self._reason)

    def add_listener(self, callback: Callable[[], None]) -> None:
        self._listeners.append(callback)
        if self._aborted:
            callback()

    def _abort(self, reason: Union[str, None] = None) -> None:
        if self._aborted:
            return
        self._aborted = True
        self._reason = reason
        self._event.set()
        for listener in self._listeners:
            try:
                listener()
            except Exception:
                pass

    async def wait(self) -> None:
        await self._event.wait()

    @staticmethod
    def timeout(seconds: float) -> "AbortSignal":
        controller = AbortController()
        async def _timeout() -> None:
            await asyncio.sleep(seconds)
            controller.abort(f"Timeout after {seconds}s")
        asyncio.create_task(_timeout())
        return controller.signal


class AbortController:
    def __init__(self) -> None:
        self._signal = AbortSignal()

    @property
    def signal(self) -> AbortSignal:
        return self._signal

    def abort(self, reason: Union[str, None] = None) -> None:
        self._signal._abort(reason)



import json
from typing import Any, List, Optional

class RobustEncoder(json.JSONEncoder):
    """Encodes Python types that aren't natively JSON serializable."""
    def default(self, obj):
        if isinstance(obj, set):
            return {"__ai_query_type__": "set", "items": list(obj)}
        return super().default(obj)

def robust_object_hook(dct):
    """Decodes JSON back into Python types."""
    if "__ai_query_type__" in dct:
        type_name = dct["__ai_query_type__"]
        if type_name == "set":
            return set(dct["items"])
    return dct

try:
    import js
except ImportError:
    # Fallback for local testing/linting outside of Cloudflare
    js = None

from ai_query.agents.storage.base import Storage


class DurableObjectStorage(Storage):
    """Storage implementation for Cloudflare Durable Objects.
    
    This class wraps the `state.storage` API available in Durable Objects.
    It handles serialization/deserialization to JSON as the storage API
    accepts JS objects (which Python dicts bridge to).
    """

    def __init__(self, storage: Any):
        """Initialize with the Durable Object storage object (ctx.storage)."""
        self._storage = storage
        if js is None:
            pass

    async def get(self, key: str) -> Any | None:
        """Get a value from DO storage."""
        # state.storage.get returns a Promise that resolves to the value
        # or undefined if not found.
        # Python bridge handles the Promise await automatically if we are in async?
        # Actually in Pyodide/workerd, we await the JsProxy Promise.
        val = await self._storage.get(key)
        if val is None or (hasattr(val, "undefined") and val.undefined):
             return None
        
        
        try:
            return json.loads(val, object_hook=robust_object_hook)
        except (TypeError, json.JSONDecodeError):
            return val

    async def set(self, key: str, value: Any) -> None:
        """Set a value in DO storage."""
        serialized = json.dumps(value, cls=RobustEncoder)
        await self._storage.put(key, serialized)

    async def delete(self, key: str) -> None:
        """Delete a key."""
        await self._storage.delete(key)

    async def keys(self, prefix: str = "") -> List[str]:
        """List keys, optionally filtering by prefix."""
        options = js.Object.fromEntries([["prefix", prefix]]) if prefix else None
        
        # This returns a Map<string, any>
        result_map = await self._storage.list(options)
        
        return list(result_map.keys())

    def transaction(self, closure: Any) -> Any:
        """Run a closure within a storage transaction.
        
        Usage:
            await self.storage.transaction(lambda txn: ...)
        """
        return self._storage.transaction(closure)



"""SQLite storage implementation."""

from __future__ import annotations

import json
import sqlite3
from pathlib import Path
from typing import Any


class SQLiteStorage:
    """SQLite-based persistent storage.

    Stores data in a SQLite database file for persistence across restarts.

    Example:
        agent = Agent("assistant", storage=SQLiteStorage("agents.db"))
    """

    def __init__(self, path: str = "agents.db") -> None:
        self._path = Path(path)
        self._path.parent.mkdir(parents=True, exist_ok=True)
        self._conn: sqlite3.Connection | None = None
        self._init_db()

    def _init_db(self) -> None:
        conn = self._get_connection()
        conn.execute("""
            CREATE TABLE IF NOT EXISTS kv_store (
                key TEXT PRIMARY KEY,
                value TEXT NOT NULL
            )
        """)
        conn.commit()

    def _get_connection(self) -> sqlite3.Connection:
        if self._conn is None:
            self._conn = sqlite3.connect(str(self._path), check_same_thread=False)
        return self._conn

    async def get(self, key: str) -> Any | None:
        conn = self._get_connection()
        cursor = conn.execute("SELECT value FROM kv_store WHERE key = ?", (key,))
        row = cursor.fetchone()
        if row is None:
            return None
        return json.loads(row[0])

    async def set(self, key: str, value: Any) -> None:
        conn = self._get_connection()
        json_value = json.dumps(value, default=str)
        conn.execute(
            "INSERT OR REPLACE INTO kv_store (key, value) VALUES (?, ?)",
            (key, json_value),
        )
        conn.commit()

    async def delete(self, key: str) -> None:
        conn = self._get_connection()
        conn.execute("DELETE FROM kv_store WHERE key = ?", (key,))
        conn.commit()

    async def keys(self, prefix: str = "") -> list[str]:
        conn = self._get_connection()
        if not prefix:
            cursor = conn.execute("SELECT key FROM kv_store")
        else:
            cursor = conn.execute(
                "SELECT key FROM kv_store WHERE key LIKE ?",
                (f"{prefix}%",),
            )
        return [row[0] for row in cursor.fetchall()]

    def close(self) -> None:
        """Close the database connection."""
        if self._conn:
            self._conn.close()
            self._conn = None



"""Transport-agnostic Agent with event emission and optional persistence."""

from __future__ import annotations

import asyncio
import json
from dataclasses import dataclass, field
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Generic,
    TYPE_CHECKING,
    TypeVar,
    Union,
    Set,
    List,
    Dict,
    Type,
)

from ai_query.agents.storage import MemoryStorage, Storage
from ai_query.agents.websocket import Connection, ConnectionContext
from ai_query.model import LanguageModel
from ai_query.types import AbortError, AbortSignal, Message

if TYPE_CHECKING:
    from ai_query.agents.events import EventBus
    from ai_query.agents.transport import AgentTransport
    from ai_query.types import ProviderOptions, StopCondition, ToolSet
    from pydantic import BaseModel


Content = Union[str, List[Any]]

# Type alias for emit handler callback
EmitHandler = Callable[[str, Dict[str, Any], int], Awaitable[None]]

# Generic TypeVar for the Agent's state. Defaults to Dict[str, Any].
StateT = TypeVar("StateT", bound=Union[Dict[str, Any], "BaseModel"])

# Generic TypeVar for type-safe agent proxies. T can be any subclass of Agent.
T = TypeVar("T", bound="Agent")


@dataclass
class Context:
    """Context for the current operation being processed by the agent."""

    connection: Union[Connection, None] = None
    ctx: Union[ConnectionContext, None] = None


def action(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to mark an agent method as an externally callable action."""
    setattr(func, "__is_action__", True)
    return func


@dataclass
class _Envelope:
    kind: str
    payload: Any
    connection: Union[Connection, None] = None
    ctx: Union[ConnectionContext, None] = None
    future: Union[asyncio.Future, None] = None


@dataclass
class Event:
    """An emitted event with ID for replay support."""

    id: int
    type: str
    data: Dict[str, Any]

    def to_dict(self) -> Dict[str, Any]:
        return {"id": self.id, "type": self.type, "data": self.data}


class AgentCallProxy(Generic[T]):
    """A type-safe proxy for making fluent calls to another agent.

    This class enables static analysis and autocompletion for remote agent calls.
    """

    def __init__(self, agent: "Agent", target_id: str):
        self._agent = agent
        self._target_id = target_id

    def __getattr__(self, name: str) -> Callable[..., Awaitable[Any]]:
        """Captures the method name and returns an awaitable RPC function."""

        async def method_proxy(**kwargs: Any) -> Any:
            """The inner function that is awaited to perform the RPC call."""
            if self._agent._transport is None:
                raise RuntimeError(
                    "No transport configured. Agents must be managed by an AgentServer "
                    "or provided a transport manually to use call()."
                )

            # The payload format is compatible with the agent's handle_invoke method
            payload = {"method": name, "params": kwargs}
            result = await self._agent._transport.invoke(self._target_id, payload)

            if "error" in result:
                raise RuntimeError(f"Agent call failed: {result['error']}")

            return result.get("result")

        return method_proxy


class Agent(Generic[StateT]):
    """A transport-agnostic AI agent with event emission and optional persistence.

    The Agent emits events during its work. How those events are delivered to
    clients is determined by the transport layer (serve(), AgentServer, etc.).

    Example:
        class ResearchBot(Agent):
            state = {"status": "idle", "results": []} # Blueprint

            async def research(self, query: str):
                await self.update_state(status="searching")
                results = await self.search(query)
                await self.update_state(status="done", results=results)
    """

    # Class-level flag to enable event persistence for replay
    enable_event_log: bool = False

    # Internal blueprint for state (don't set this directly, use 'state = ...')
    _state_blueprint: Any = None
    state_model: Any = None

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # Magic: Sniff the 'state' attribute in the class body
        if "state" in cls.__dict__:
            blueprint = cls.__dict__["state"]
            # Check if it's a Pydantic model class
            if isinstance(blueprint, type) and (
                hasattr(blueprint, "model_validate") or hasattr(blueprint, "validate")
            ):
                cls.state_model = blueprint
                cls._state_blueprint = None
            else:
                # Treat as a dictionary/initial value blueprint
                cls._state_blueprint = blueprint
                cls.state_model = None

            # Remove it from class dict so the 'state' @property works
            # (Otherwise it shadows the property)
            try:
                delattr(cls, "state")
            except (AttributeError, KeyError):
                pass

    def __init__(
        self,
        id: str,
        *,
        model: Union[LanguageModel, None] = None,
        system: str = "You are a helpful assistant.",
        tools: Union[ToolSet, None] = None,
        storage: Union[Storage, None] = None,
        initial_state: Union[StateT, None] = None,
        stop_when: Union[StopCondition, List[StopCondition], None] = None,
        provider_options: Union[ProviderOptions, None] = None,
        transport: Union[AgentTransport, None] = None,
        event_bus: Union[EventBus, None] = None,
    ) -> None:
        self._id = id
        self._storage = storage or MemoryStorage()

        # Combine class-level blueprint with instance-level initial_state
        if initial_state is not None:
            self._initial_state = initial_state
        else:
            # Fallback to blueprint if it's a dict
            self._initial_state = (
                self._state_blueprint if self._state_blueprint is not None else {}
            )

        self.model = model
        self.system = system
        self.tools = tools or {}
        self.stop_when = stop_when
        self.provider_options = provider_options

        # Agent state
        self._state: Union[StateT, None] = None
        self._messages: List[Message] = []

        # Event log for durability & replay
        self._event_log: List[Dict[str, Any]] = []
        self._event_counter: int = 0

        # Communication
        self._transport = transport
        self._event_bus = event_bus

        # Emit handler - injected by transport layer (serve, AgentServer, etc.)
        # Signature: async def handler(event_type: str, data: dict, event_id: int)
        self._emit_handler: Union[EmitHandler, None] = None

        # Connection tracking (used by serve() internally)
        self._connections: Set[Connection] = set()
        self._sse_connections: Set[Any] = set()

        # Actor mailbox for serialized message processing
        self._mailbox: asyncio.Queue[_Envelope] = asyncio.Queue()
        self._processor_task: Union[asyncio.Task, None] = None
        self._running = False
        self._start_lock = asyncio.Lock()

        # Current context (only valid during message processing)
        self.context = Context()

        # Action registry
        self._action_map: Dict[str, Callable[..., Awaitable[Any]]] = {}

    # ─── Properties ────────────────────────────────────────────────────────

    @property
    def id(self) -> str:
        return self._id

    @property
    def storage(self) -> Storage:
        return self._storage

    @property
    def state(self) -> StateT:
        if self._state is None:
            raise RuntimeError(
                "Agent not started. Call 'await agent.start()' or use "
                "'async with agent:' context manager."
            )
        return self._state

    @property
    def messages(self) -> List[Message]:
        return self._messages

    # ─── Event Emission (The Core API) ─────────────────────────────────────

    async def emit(self, event: str, data: Dict[str, Any]) -> int:
        """Emit an event. Returns the event ID."""
        # Assign ID
        self._event_counter += 1
        event_id = self._event_counter

        # Log event
        event_record = {"id": event_id, "type": event, "data": data}
        self._event_log.append(event_record)

        # Persist if enabled
        if self.enable_event_log:
            await self._storage.set(f"{self._id}:event_log", self._event_log)

        # Deliver via handler (set by transport layer)
        if self._emit_handler:
            await self._emit_handler(event, data, event_id)

        return event_id

    async def replay_events(self, after_id: int = 0) -> AsyncIterator[Event]:
        """Yield events after the given ID for replay."""
        for event_record in self._event_log:
            if event_record["id"] > after_id:
                yield Event(
                    id=event_record["id"],
                    type=event_record["type"],
                    data=event_record["data"],
                )

    async def clear_event_log(self) -> None:
        """Clear the event log."""
        self._event_log = []
        self._event_counter = 0
        if self.enable_event_log:
            await self._storage.delete(f"{self._id}:event_log")

    # ─── State Management ──────────────────────────────────────────────────

    async def set_state(self, state: StateT) -> None:
        """Set the agent's state and persist to storage."""
        self._state = state
        await self.save_state()
        # Emit state change event
        payload = self._get_state_dict()
        await self.emit("state", {"state": payload})

    async def update_state(self, **kwargs: Any) -> None:
        """Merge updates into the agent's state."""
        if hasattr(self.state, "model_copy"):
            # Pydantic v2
            new_state = self.state.model_copy(update=kwargs)  # type: ignore
        elif hasattr(self.state, "copy"):
            # Dict or Pydantic v1
            if isinstance(self.state, dict):
                new_state = {**self.state, **kwargs}  # type: ignore
            else:
                # Pydantic v1
                new_state = self.state.copy(update=kwargs)  # type: ignore
        else:
            raise TypeError(f"Unsupported state type: {type(self.state)}")

        await self.set_state(new_state)  # type: ignore

    async def save_state(self) -> None:
        """Persist the current state to storage."""
        payload = self._get_state_dict()
        await self._storage.set(f"{self._id}:state", payload)

    def _get_state_dict(self) -> Dict[str, Any]:
        """Convert state to a dictionary for storage or transit."""
        if self._state is None:
            return {}
        if hasattr(self._state, "model_dump"):
            return self._state.model_dump()  # type: ignore
        if hasattr(self._state, "dict"):
            return self._state.dict()  # type: ignore
        return self._state  # type: ignore

    async def clear(self) -> None:
        """Clear message history."""
        self._messages = []
        await self._storage.set(f"{self._id}:messages", [])

    # ─── Chat & Streaming ──────────────────────────────────────────────────

    async def chat(
        self,
        message: Content,
        *,
        signal: Union[AbortSignal, None] = None,
    ) -> str:
        """Send a message and get a response."""
        if signal:
            signal.throw_if_aborted()

        if self.model is None:
            raise ValueError(
                "No model provided for Agent. Please pass a model to the Agent constructor "
                "or set agent.model before calling chat()."
            )

        msg_content = message if isinstance(message, str) else message
        self._messages.append(Message(role="user", content=msg_content))

        if signal:
            gen_task = asyncio.create_task(self._do_chat())
            abort_task = asyncio.create_task(signal.wait())

            done, pending = await asyncio.wait(
                [gen_task, abort_task],
                return_when=asyncio.FIRST_COMPLETED,
            )

            for task in pending:
                task.cancel()

            if abort_task in done:
                self._messages.pop()
                raise AbortError(signal.reason)

            return gen_task.result()
        else:
            return await self._do_chat()

    async def _do_chat(self) -> str:
        from ai_query import stream_text

        # Ensure model is set (should have been checked by caller)
        if self.model is None:
            raise ValueError("Agent model is not configured")

        result = stream_text(
            model=self.model,
            system=self.system,
            messages=self._messages,
            tools=self.tools if self.tools else None,
            stop_when=self.stop_when,
            provider_options=self.provider_options,
        )

        full_response = ""
        async for chunk in result.text_stream:
            full_response += chunk

        self._messages.append(Message(role="assistant", content=full_response))
        await self._storage.set(
            f"{self._id}:messages", [m.to_dict() for m in self._messages]
        )
        return full_response

    async def stream(
        self,
        message: Content,
        *,
        signal: Union[AbortSignal, None] = None,
    ) -> AsyncIterator[str]:
        """Stream a response chunk by chunk."""
        if signal:
            signal.throw_if_aborted()

        if self.model is None:
            raise ValueError(
                "No model provided for Agent. Please pass a model to the Agent constructor "
                "or set agent.model before calling chat()."
            )

        # Ensure model is set for type checker
        assert self.model is not None

        msg_content = message if isinstance(message, str) else message
        self._messages.append(Message(role="user", content=msg_content))

        from ai_query import stream_text

        result = stream_text(
            model=self.model,
            system=self.system,
            messages=self._messages,
            tools=self.tools if self.tools else None,
            stop_when=self.stop_when,
            provider_options=self.provider_options,
        )

        full_response = ""
        async for chunk in result.text_stream:
            if signal and signal.aborted:
                self._messages.pop()
                raise AbortError(signal.reason)
            full_response += chunk
            yield chunk

        self._messages.append(Message(role="assistant", content=full_response))
        await self._storage.set(
            f"{self._id}:messages", [m.to_dict() for m in self._messages]
        )

    # ─── Lifecycle ─────────────────────────────────────────────────────────

    async def start(self) -> None:
        """Start the agent, loading state from storage.

        This method is idempotent and concurrency-safe.
        """
        # Hot State Optimization: Check if already started
        if self._state is not None:
            return

        async with self._start_lock:
            # Double-check inside lock
            if self._state is not None:
                return

            # Load state
            state_data = await self._storage.get(f"{self._id}:state")

            if state_data is not None:
                if self.state_model:
                    # Use Pydantic to parse
                    if hasattr(self.state_model, "model_validate"):
                        self._state = self.state_model.model_validate(state_data)  # type: ignore
                    else:
                        self._state = self.state_model(**state_data)  # type: ignore
                else:
                    self._state = state_data  # type: ignore
            else:
                # Use initial state
                initial = self._initial_state if self._initial_state is not None else {}
                if self.state_model and isinstance(initial, dict):
                    if hasattr(self.state_model, "model_validate"):
                        self._state = self.state_model.model_validate(initial)  # type: ignore
                    else:
                        self._state = self.state_model(**initial)  # type: ignore
                else:
                    self._state = initial  # type: ignore

            # Load messages
            messages_data = await self._storage.get(f"{self._id}:messages")
            if messages_data:
                self._messages = [
                    Message(role=m["role"], content=m["content"]) for m in messages_data
                ]

            # Load event log if persistence is enabled
            if self.enable_event_log:
                event_log_data = await self._storage.get(f"{self._id}:event_log")
                if event_log_data:
                    self._event_log = event_log_data
                    self._event_counter = max(
                        (e.get("id", 0) for e in self._event_log), default=0
                    )

            # Build action map
            for name in dir(self):
                attr = getattr(self, name)
                if hasattr(attr, "__is_action__"):
                    self._action_map[name] = attr

            self._running = True
            self._processor_task = asyncio.create_task(self._process_mailbox())
            await self.on_start()

    async def stop(self) -> None:
        """Stop the agent."""
        self._running = False
        if self._processor_task is not None:
            self._processor_task.cancel()
            try:
                await self._processor_task
            except asyncio.CancelledError:
                pass
            self._processor_task = None
        await self.on_stop()

    async def on_start(self) -> None:
        """Override to run logic when agent starts."""
        pass

    async def on_stop(self) -> None:
        """Override to run logic when agent stops."""
        pass

    async def __aenter__(self) -> "Agent":
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.stop()

    # ─── Connection Handling (used by serve/transport) ─────────────────────

    async def on_connect(self, connection: Connection, ctx: ConnectionContext) -> None:
        """Called when a client connects. Override for custom logic."""
        self._connections.add(connection)

    async def on_message(
        self, connection: Connection, message: Union[str, bytes]
    ) -> None:
        """Called when a message is received. Override to handle messages."""
        pass

    async def on_close(self, connection: Connection, code: int, reason: str) -> None:
        """Called when a client disconnects. Override for cleanup."""
        self._connections.discard(connection)

    # ─── Actor Mailbox ─────────────────────────────────────────────────────

    async def _process_mailbox(self) -> None:
        """Process enqueued messages serially (actor pattern)."""
        while self._running:
            try:
                envelope = await self._mailbox.get()
            except asyncio.CancelledError:
                break

            try:
                result = await self._handle_envelope(envelope)
                if envelope.future is not None and not envelope.future.done():
                    envelope.future.set_result(result)
            except Exception as e:
                if envelope.future is not None and not envelope.future.done():
                    envelope.future.set_exception(e)
            finally:
                self._mailbox.task_done()

    async def _handle_envelope(self, envelope: _Envelope) -> Any:
        # Set context for this operation
        self.context.connection = envelope.connection
        self.context.ctx = envelope.ctx

        try:
            if envelope.kind == "connect":
                assert envelope.connection is not None
                assert envelope.ctx is not None
                await self.on_connect(envelope.connection, envelope.ctx)
            elif envelope.kind == "message":
                assert envelope.connection is not None
                await self.on_message(envelope.connection, envelope.payload)
            elif envelope.kind == "close":
                assert envelope.connection is not None
                code, reason = envelope.payload
                await self.on_close(envelope.connection, code, reason)
            elif envelope.kind == "action":
                name = envelope.payload["name"]
                params = envelope.payload["params"]
                if name not in self._action_map:
                    raise ValueError(f"Action '{name}' not found on agent {self.id}")

                res = self._action_map[name](**params)
                if asyncio.iscoroutine(res):
                    return await res
                return res
            return None
        finally:
            # Clear context
            self.context.connection = None
            self.context.ctx = None

    def enqueue(
        self,
        kind: str,
        payload: Any = None,
        connection: Union[Connection, None] = None,
        ctx: Union[ConnectionContext, None] = None,
        future: Union[asyncio.Future, None] = None,
    ) -> None:
        """Enqueue a message for serial processing."""
        self._mailbox.put_nowait(
            _Envelope(
                kind=kind,
                payload=payload,
                connection=connection,
                ctx=ctx,
                future=future,
            )
        )

    # ─── Server ────────────────────────────────────────────────────────────

    def serve(
        self,
        host: str = "localhost",
        port: int = 8080,
    ) -> None:
        """Start a server with WebSocket, SSE, and REST endpoints."""
        from ai_query.agents.server import AgentServer

        server = AgentServer(self.__class__)
        server.serve(host=host, port=port)

    # ─── Serverless Request Handling ───────────────────────────────────────

    async def handle_request(self, request: Dict[str, Any]) -> Dict[str, Any]:
        """Handle a serverless request."""
        if self._state is None:
            await self.start()

        action = request.get("action", "chat")

        if action == "chat":
            message = request.get("message", "")
            response = await self.chat(message)
            return {"agent_id": self.id, "response": response}

        elif action == "action":
            name = request.get("name")
            params = request.get("params", {})
            if not name:
                return {"error": "Missing 'name' in action request"}

            # Enqueue to ensure sequential execution
            future = asyncio.get_running_loop().create_future()
            self.enqueue("action", {"name": name, "params": params}, future=future)
            result = await future
            return {"agent_id": self.id, "result": result}

        elif action == "invoke":
            payload = request.get("payload", {})
            # Try to handle as action if it has 'method'
            if "method" in payload:
                method = payload["method"]
                params = payload.get("params", {})
                if method in self._action_map:
                    future = asyncio.get_running_loop().create_future()
                    self.enqueue(
                        "action", {"name": method, "params": params}, future=future
                    )
                    result = await future
                    return {"agent_id": self.id, "result": result}

            result = await self.handle_invoke(payload)
            return {"agent_id": self.id, "result": result}

        elif action == "state":
            return {"agent_id": self.id, "state": self._get_state_dict()}

        return {"error": f"Unknown action: {action}"}

    def call(self, agent_id: str, *, agent_cls: type[T]) -> AgentCallProxy[T]:
        """Returns a type-safe proxy for making fluent calls to another agent."""
        return AgentCallProxy(self, agent_id)

    async def handle_invoke(self, payload: Dict[str, Any]) -> Dict[str, Any]:
        """Handle an incoming call from another agent."""
        method = payload.get("method")
        params = payload.get("params", {})

        if not method:
            return {"error": "Missing 'method' in call payload"}

        # 1. Check if it's a registered action
        if method in self._action_map:
            try:
                # Enqueue to ensure sequential execution
                future = asyncio.get_running_loop().create_future()
                self.enqueue(
                    "action", {"name": method, "params": params}, future=future
                )
                result = await future
                return {"result": result}
            except Exception as e:
                return {"error": str(e)}

        # 2. Legacy support: look for on_call_ method on self
        handler = getattr(self, f"on_call_{method}", None)
        if handler and callable(handler):
            try:
                res = handler(**params)
                if asyncio.iscoroutine(res):
                    result = await res
                else:
                    result = res
                return {"result": result}
            except Exception as e:
                return {"error": str(e)}

        return {"error": f"Method '{method}' not implemented on agent {self.id}"}

    async def handle_request_stream(
        self, request: Dict[str, Any]
    ) -> AsyncIterator[str]:
        """Serverless streaming request handler."""
        # Ensure agent is started
        if self._state is None:
            await self.start()

        action = request.get("action", "chat")

        if action == "chat":
            message = request.get("message", "")

            try:
                # Start event
                yield "event: start\ndata: \n\n"

                full_text = ""
                # Stream chunks using the agent's stream method
                async for chunk in self.stream(message):
                    full_text += chunk
                    # SSE format: escape newlines for safety
                    safe_chunk = chunk.replace("\n", "\\n")
                    yield f"event: chunk\ndata: {safe_chunk}\n\n"

                # End event with full text
                safe_full = json.dumps(full_text)
                yield f"event: end\ndata: {safe_full}\n\n"

            except Exception as e:
                yield f"event: error\ndata: {str(e)}\n\n"
        else:
            yield f"event: error\ndata: Streaming not supported for action: {action}\n\n"



"""Battle Test: End-to-end multi-agent workflow."""

import asyncio
import os
from ai_query import Agent, action, AgentRegistry, AgentServer, HTTPTransport, connect, generate_text
from ai_query.providers import google

# Get API key from environment
# GOOGLE_API_KEY = os.getenv("GOOGLE_API_KEY")
# if not GOOGLE_API_KEY:
#     raise ValueError("Please set GOOGLE_API_KEY environment variable")

# --- Agent Definitions ---

class ResearcherAgent(Agent):
    """Researches a topic using a real LLM."""
    @action
    async def research(self, topic: str) -> str:
        print(f"[{self.id}] Researching topic: {topic}")
        res = await generate_text(
            model=google("gemini-2.5-flash", api_key=GOOGLE_API_KEY),
            prompt=f"Provide a concise, 3-sentence summary on the topic: {topic}"
        )
        print(f"[{self.id}] Research complete.")
        return res.text

class WriterAgent(Agent):
    """Writes a summary based on research."""
    @action
    async def write(self, research_summary: str) -> str:
        print(f"[{self.id}] Writing content based on research...")
        res = await generate_text(
            model=google("gemini-2.5-flash", api_key=GOOGLE_API_KEY),
            prompt=f"Write a short blog post intro based on this summary:\n\n{research_summary}"
        )
        print(f"[{self.id}] Writing complete.")
        return res.text

class CoordinatorAgent(Agent):
    """Orchestrates the research and writing workflow."""
    @action
    async def create_article(self, topic: str) -> dict:
        print(f"[{self.id}] Starting workflow for topic: {topic}")
        
        # Call the remote researcher
        researcher_proxy = self.call("researcher", agent_cls=ResearcherAgent)
        research_summary = await researcher_proxy.research(topic=topic)
        print(f"[{self.id}] Got research summary: '{research_summary[:30]}...'")
        
        # Call the remote writer
        writer_proxy = self.call("writer", agent_cls=WriterAgent)
        article_intro = await writer_proxy.write(research_summary=research_summary)
        print(f"[{self.id}] Got article intro: '{article_intro[:30]}...'")
        
        return {
            "topic": topic,
            "research": research_summary,
            "article": article_intro
        }

# --- Remote Server (Simulates a separate microservice) ---

def run_remote_server():
    """Runs the remote agent server."""
    print("Starting remote server on port 8091...")
    registry = AgentRegistry()
    registry.register("researcher", ResearcherAgent)
    registry.register("writer", WriterAgent)
    
    server = AgentServer(registry)
    server.serve(port=8091)

# --- Main Server (Runs the coordinator) ---

def run_main_server():
    """Runs the main coordinator agent server."""
    print("Starting main server on port 8090...")
    registry = AgentRegistry()
    registry.register("coordinator", CoordinatorAgent)
    
    # Register the remote agents so the coordinator knows where to find them
    remote_transport = HTTPTransport(base_url="http://localhost:8091/agent")
    registry.register("researcher", remote_transport)
    registry.register("writer", remote_transport)
    
    server = AgentServer(registry)
    server.serve(port=8090)

# --- Test Client ---

async def run_test_client():
    """Connects to the main server and runs the workflow."""
    print("\n--- Running Battle Test Client ---")
    await asyncio.sleep(2) # Wait for servers to start
    
    try:
        coordinator = connect("http://localhost:8090/agent/coordinator")
        
        topic = "The future of artificial intelligence in education"
        print(f"Client: Requesting article on '{topic}'...")
        
        result = await coordinator.call().create_article(topic=topic)
        
        print("\n--- Battle Test SUCCESS ---")
        print(f"Final Article Intro:\n{result['article']}")
        
        assert "education" in result["article"].lower()
        
    except Exception as e:
        print(f"\n--- Battle Test FAILED ---")
        print(f"Error: {e}")
    finally:
        # In a real script, we'd have a way to shut down the servers.
        # For this test, we'll just exit.
        print("\n--- Battle Test Complete ---")
        # This is a bit abrupt, but necessary for an automated script.
        import os
        os._exit(0)


# if __name__ == "__main__":
#     import multiprocessing
    
#     # Run servers in separate processes
#     p_remote = multiprocessing.Process(target=run_remote_server)
#     p_main = multiprocessing.Process(target=run_main_server)
    
#     p_remote.start()
#     p_main.start()
    
#     # Run client
#     asyncio.run(run_test_client())



from __future__ import annotations

import asyncio
import json
from typing import Any

import aiohttp
import pytest
import pytest_asyncio
from aiohttp import web

from ai_query.agents import Agent, action, AgentServer, MemoryStorage


class RpcAgent(Agent):
    def __init__(self, id: str, **kwargs: Any):
        super().__init__(
            id,
            storage=MemoryStorage(),
            initial_state={"reputation": {}},
            **kwargs
        )

    @action
    async def get_id(self) -> str:
        return self.id

    @action
    async def get_state(self) -> dict[str, Any]:
        return self.state

    @action
    async def update_reputation(self, amount: int) -> int:
        assert self.context.connection is not None
        user_id = self.context.connection.state.get("user_id", "anonymous")
        reputation = self.state.setdefault("reputation", {})
        reputation[user_id] = reputation.get(user_id, 0) + amount
        await self.update_state(reputation=reputation)
        return reputation[user_id]

    async def on_connect(self, connection, ctx):
        connection.state["user_id"] = ctx.metadata.get("user_id", "anonymous")
        await super().on_connect(connection, ctx)


@pytest_asyncio.fixture
async def rpc_server():
    server = AgentServer(RpcAgent)
    app = server.create_app()
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, "localhost", 8080)
    await site.start()
    yield
    await runner.cleanup()


@pytest.mark.asyncio
async def test_http_action_call(rpc_server):
    async with aiohttp.ClientSession() as session:
        # Test basic action
        async with session.post(
            "http://localhost:8080/agent/test-agent/action/get_id"
        ) as resp:
            assert resp.status == 200
            data = await resp.json()
            assert data["result"] == "test-agent"

        # Test action with no params
        async with session.post(
            "http://localhost:8080/agent/test-agent/action/get_state"
        ) as resp:
            assert resp.status == 200
            data = await resp.json()
            assert data["result"] == {"reputation": {}}


@pytest.mark.asyncio
async def test_websocket_action_call(rpc_server):
    async with aiohttp.ClientSession() as session:
        async with session.ws_connect(
            "http://localhost:8080/agent/ws-agent/ws?user_id=alice"
        ) as ws:
            # Call action with params
            call_id = "123"
            await ws.send_json(
                {
                    "type": "action",
                    "name": "update_reputation",
                    "params": {"amount": 10},
                    "call_id": call_id,
                }
            )

            # Wait for result, ignoring other messages until we get the one we want
            response = {}
            for _ in range(5): # Try up to 5 times
                response = await ws.receive_json()
                if response.get("type") == "action_result" and response.get("call_id") == call_id:
                    break
            
            assert response["type"] == "action_result"
            assert response["call_id"] == call_id
            assert response["result"] == 10

            # Check state
            async with session.get(
                "http://localhost:8080/agent/ws-agent/state"
            ) as resp:
                data = await resp.json()
                assert data["reputation"]["alice"] == 10



"""FastAPI adapter for serving agents."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

try:
    from fastapi import APIRouter, Request, HTTPException
    from fastapi.responses import JSONResponse, StreamingResponse
    from pydantic import BaseModel
except ImportError:
    # Allow import without optional dependencies
    APIRouter = object # type: ignore

if TYPE_CHECKING:
    from ai_query.agents.agent import Agent
    from ai_query.agents.registry import AgentRegistry


class AgentRouter(APIRouter):
    """FastAPI Router for serving Agents.
    
    Supports serving a single Agent instance or an AgentRegistry for multi-agent routing.

    Endpoints (Single Agent):
    - POST /chat
    - POST /invoke
    - GET /state
    - POST /

    Endpoints (Registry):
    - POST /{agent_id}/chat
    - POST /{agent_id}/invoke
    - GET /{agent_id}/state
    - POST /{agent_id}
    
    Example:
        # Single Agent
        app.include_router(AgentRouter(my_agent), prefix="/agent/bot")

        # Registry (Multi-Agent)
        registry = AgentRegistry()
        registry.register("worker-.*", WorkerAgent)
        app.include_router(AgentRouter(registry), prefix="/agents")
    """

    def __init__(self, target: Union["Agent", "AgentRegistry"], **kwargs: Any):
        if APIRouter is object:
            raise ImportError("FastAPI is not installed. Run 'pip install fastapi'.")
            
        super().__init__(**kwargs)
        
        from ai_query.agents.registry import AgentRegistry
        
        self.is_registry = isinstance(target, AgentRegistry)
        self.target = target
        
        if self.is_registry:
            # Register routes with agent_id path param
            self.add_api_route("/{agent_id}/chat", self._handle_chat, methods=["POST"])
            self.add_api_route("/{agent_id}/invoke", self._handle_invoke, methods=["POST"])
            self.add_api_route("/{agent_id}/state", self._handle_get_state, methods=["GET"])
            self.add_api_route("/{agent_id}", self._handle_request, methods=["POST"])
        else:
            # Register routes at root
            self.add_api_route("/chat", self._handle_chat, methods=["POST"])
            self.add_api_route("/invoke", self._handle_invoke, methods=["POST"])
            self.add_api_route("/state", self._handle_get_state, methods=["GET"])
            self.add_api_route("/", self._handle_request, methods=["POST"])

    async def _get_agent(self, agent_id: Union[str, None] = None) -> "Agent":
        if self.is_registry:
            if not agent_id:
                raise HTTPException(status_code=400, detail="Missing agent_id")
            try:
                # Resolve returns class or transport
                target = self.target.resolve(agent_id)
                # Instantiate if class
                if isinstance(target, type):
                    return target(agent_id)
                # If transport, wrap in RemoteAgent? 
                # Ideally we shouldn't be routing to remote agents via this router 
                # unless acting as a gateway. For now assuming local execution.
                raise HTTPException(status_code=500, detail="Remote routing not supported in AgentRouter yet")
            except ValueError:
                raise HTTPException(status_code=404, detail=f"Agent '{agent_id}' not found")
        else:
            return self.target

    async def _ensure_started(self, agent: "Agent") -> None:
        if agent._state is None:
            await agent.start()

    async def _handle_chat(self, request: Request, agent_id: Union[str, None] = None) -> Any:
        agent = await self._get_agent(agent_id)
        await self._ensure_started(agent)
        
        try:
            body = await request.json()
        except Exception:
            raise HTTPException(status_code=400, detail="Invalid JSON")
            
        message = body.get("message", "")
        if not message:
            raise HTTPException(status_code=400, detail="Missing 'message'")
            
        # Check for streaming
        if request.query_params.get("stream", "").lower() == "true":
            return StreamingResponse(
                self._stream_generator(agent, message),
                media_type="text/event-stream"
            )
            
        result = await agent.handle_request({
            "action": "chat", 
            "message": message
        })
        return JSONResponse(result)

    async def _stream_generator(self, agent: "Agent", message: str) -> Any:
        stream_req = {"action": "chat", "message": message}
        async for chunk in agent.handle_request_stream(stream_req):
            yield chunk

    async def _handle_invoke(self, request: Request, agent_id: Union[str, None] = None) -> Any:
        agent = await self._get_agent(agent_id)
        await self._ensure_started(agent)
        
        try:
            body = await request.json()
        except Exception:
            raise HTTPException(status_code=400, detail="Invalid JSON")
            
        payload = body.get("payload", body)
        result = await agent.handle_request({
            "action": "invoke",
            "payload": payload
        })
        return JSONResponse(result)

    async def _handle_get_state(self, agent_id: Union[str, None] = None) -> Any:
        agent = await self._get_agent(agent_id)
        await self._ensure_started(agent)
        try:
            return JSONResponse(agent.state)
        except Exception:
            raise HTTPException(status_code=500, detail="State serialization failed")

    async def _handle_request(self, request: Request, agent_id: Union[str, None] = None) -> Any:
        agent = await self._get_agent(agent_id)
        await self._ensure_started(agent)
        
        try:
            body = await request.json()
        except Exception:
            raise HTTPException(status_code=400, detail="Invalid JSON")
            
        result = await agent.handle_request(body)
        return JSONResponse(result)
